# -*- coding: utf-8 -*-


F_vmap_fx = 'F_vmap_fx'
F_vmap_fy = 'F_vmap_fy'
F_vmap_brentq_fx = 'F_vmap_brentq_fx'
F_vmap_brentq_fy = 'F_vmap_brentq_fy'
F_vmap_fp_aux = 'vmap_fixed_point_aux'
F_vmap_fp_opt = 'vmap_fixed_point_opt'
F_vmap_dfxdx = 'F_vmap_dfxdx'
F_fx = 'F_fx'
F_fy = 'F_fy'
F_fz = 'F_fz'
F_dfxdx = 'F_dfxdx'
F_dfxdy = 'F_dfxdy'
F_dfydx = 'F_dfydx'
F_dfydy = 'F_dfydy'
F_jacobian = 'F_jacobian'
F_vmap_jacobian = 'F_vmap_jacobian'
F_fixed_point = 'F_fixed_point'
F_fixed_point_aux = 'F_fixed_point_aux'
F_fixed_point_opt = 'F_fixed_point_opt'
F_fx_nullcline_by_opt = 'F_fx_nullcline_by_opt'
F_fy_nullcline_by_opt = 'F_fy_nullcline_by_opt'
F_x_in_all = 'F_x_in_all'
F_y_in_all = 'F_y_in_all'
F_x_by_y = 'F_x_by_y'
F_y_by_x = 'F_y_by_x'
F_y_convert = 'F_y_convert'
F_x_convert = 'F_x_convert'
F_int_x = 'F_int_x'
F_int_y = 'F_int_y'

x_by_y = 'x_by_y'
y_by_x = 'y_by_x'
y_by_x_in_fy = 'fy::y=f(x)'
y_by_x_in_fx = 'fx::y=f(x)'
x_by_y_in_fx = 'fx::x=f(y)'
x_by_y_in_fy = 'fy::x=f(y)'
F_y_by_x_in_fy = 'F[fy::y=f(x)]'
F_x_by_y_in_fy = 'F[fy::x=f(y)]'
F_y_by_x_in_fx = 'F[fx::y=f(x)]'
F_x_by_y_in_fx = 'F[fx::x=f(y)]'
fx_nullcline_points = 'fx_nullcline_points'
fy_nullcline_points = 'fy_nullcline_points'
sympy_failed = 'sympy_failed'
sympy_success = 'sympy_success'
sympy_escape = 'sympy_escape'
sympy_timeout = 'sympy_timeout'
fx_sign = 'fx_sign'
fy_sign = 'fy_sign'
par_eval_parallel = 'par_eval_parallel'
par_eval_iter = 'par_eval_iter'
prefix = '\t'




# -*- coding: utf-8 -*-

import jax.numpy.fft

from brainpy.math.jaxarray import JaxArray
from brainpy.math.numpy_ops import _remove_jaxarray

__all__ = [
  "fft", "fft2", "fftfreq", "fftn", "fftshift", "hfft",
  "ifft", "ifft2", "ifftn", "ifftshift", "ihfft", "irfft",
  "irfft2", "irfftn", "rfft", "rfft2", "rfftfreq", "rfftn"
]


def fft(a, n=None, axis=-1, norm=None):
  a = _remove_jaxarray(a)
  return JaxArray(jax.numpy.fft.fft(a=a, n=n, axis=axis, norm=norm))


def fft2(a, s=None, axes=(-2, -1), norm=None):
  a = _remove_jaxarray(a)
  return JaxArray(jax.numpy.fft.fft2(a=a, s=s, axes=axes, norm=norm))


def fftfreq(n, d=1.0):
  return JaxArray(jax.numpy.fft.fftfreq(n=n, d=d))


def fftn(a, s=None, axes=None, norm=None):
  a = _remove_jaxarray(a)
  return JaxArray(jax.numpy.fft.fftn(a=a, s=s, axes=axes, norm=norm))


def fftshift(x, axes=None):
  x = _remove_jaxarray(x)
  return JaxArray(jax.numpy.fft.fftshift(x=x, axes=axes))


def hfft(a, n=None, axis=-1, norm=None):
  a = _remove_jaxarray(a)
  return JaxArray(jax.numpy.fft.hfft(a=a, n=n, axis=axis, norm=norm))


def ifft(a, n=None, axis=-1, norm=None):
  a = _remove_jaxarray(a)
  return JaxArray(jax.numpy.fft.ifft(a=a, n=n, axis=axis, norm=norm))


def ifft2(a, s=None, axes=(-2, -1), norm=None):
  a = _remove_jaxarray(a)
  return JaxArray(jax.numpy.fft.ifft2(a=a, s=s, axes=axes, norm=norm))


def ifftn(a, s=None, axes=None, norm=None):
  a = _remove_jaxarray(a)
  return JaxArray(jax.numpy.fft.ifftn(a=a, s=s, axes=axes, norm=norm))


def ifftshift(x, axes=None):
  x = _remove_jaxarray(x)
  return JaxArray(jax.numpy.fft.ifftshift(x=x, axes=axes))


def ihfft(a, n=None, axis=-1, norm=None):
  a = _remove_jaxarray(a)
  return JaxArray(jax.numpy.fft.ihfft(a=a, n=n, axis=axis, norm=norm))


def irfft(a, n=None, axis=-1, norm=None):
  a = _remove_jaxarray(a)
  return JaxArray(jax.numpy.fft.irfft(a=a, n=n, axis=axis, norm=norm))


def irfft2(a, s=None, axes=(-2, -1), norm=None):
  a = _remove_jaxarray(a)
  return JaxArray(jax.numpy.fft.irfft2(a=a, s=s, axes=axes, norm=norm))


def irfftn(a, s=None, axes=None, norm=None):
  a = _remove_jaxarray(a)
  return JaxArray(jax.numpy.fft.irfftn(a=a, s=s, axes=axes, norm=norm))


def rfft(a, n=None, axis=-1, norm=None):
  a = _remove_jaxarray(a)
  return JaxArray(jax.numpy.fft.rfft(a=a, n=n, axis=axis, norm=norm))


def rfft2(a, s=None, axes=(-2, -1), norm=None):
  a = _remove_jaxarray(a)
  return JaxArray(jax.numpy.fft.rfft2(a=a, s=s, axes=axes, norm=norm))


def rfftfreq(n, d=1.0):
  return JaxArray(jax.numpy.fft.rfftfreq(n=n, d=d))


def rfftn(a, s=None, axes=None, norm=None):
  a = _remove_jaxarray(a)
  return JaxArray(jax.numpy.fft.rfftn(a=a, s=s, axes=axes, norm=norm))



# -*- coding: utf-8 -*-

import logging

import numpy as np

from brainpy.errors import ConnectorError

from .base import *

logger = logging.getLogger('brainpy.building.connect')

__all__ = [
  'One2One', 'one2one',
  'All2All', 'all2all',
  'GridFour', 'grid_four',
  'GridEight', 'grid_eight',
  'GridN',
]


class One2One(TwoEndConnector):
  """Connect two neuron groups one by one. This means
  The two neuron groups should have the same size.
  """
  def __init__(self):
    super(One2One, self).__init__()

  def __call__(self, pre_size, post_size):
    super(One2One, self).__call__(pre_size, post_size)
    try:
      assert self.pre_num == self.post_num
    except AssertionError:
      raise ConnectorError(f'One2One connection must be defined in two groups with the '
                           f'same size, but {self.pre_num} != {self.post_num}.')
    return self

  def build_conn(self):
    ind = np.arange(self.pre_num)
    indptr = np.arange(self.pre_num + 1)

    return dict(csr=(ind, indptr), mat=None, ij=None)


one2one = One2One()


class All2All(TwoEndConnector):
  """Connect each neuron in first group to all neurons in the
  post-synaptic neuron groups. It means this kind of conn
  will create (num_pre x num_post) synapses.
  """

  def __init__(self, include_self=True):
    self.include_self = include_self
    super(All2All, self).__init__()

  def build_conn(self):
    mat = np.ones((self.pre_num, self.post_num), dtype=MAT_DTYPE)
    if not self.include_self:
      np.fill_diagonal(mat, False)

    return dict(csr=None, mat=mat, ij=None)


all2all = All2All(include_self=True)


# @tools.numba_jit
def _grid_four(height, width, row, include_self):
  conn_i = []
  conn_j = []

  for col in range(width):
    i_index = (row * width) + col
    if 0 <= row - 1 < height:
      j_index = ((row - 1) * width) + col
      conn_i.append(i_index)
      conn_j.append(j_index)
    if 0 <= row + 1 < height:
      j_index = ((row + 1) * width) + col
      conn_i.append(i_index)
      conn_j.append(j_index)
    if 0 <= col - 1 < width:
      j_index = (row * width) + col - 1
      conn_i.append(i_index)
      conn_j.append(j_index)
    if 0 <= col + 1 < width:
      j_index = (row * width) + col + 1
      conn_i.append(i_index)
      conn_j.append(j_index)
    if include_self:
      conn_i.append(i_index)
      conn_j.append(i_index)
  return conn_i, conn_j


class GridFour(OneEndConnector):
  """The nearest four neighbors conn method."""

  def __init__(self, include_self=False):
    super(GridFour, self).__init__()
    self.include_self = include_self

  def build_conn(self):
    # only the 1- or 2-D structure is supported
    if len(self.pre_size) == 1:
      height, width = self.pre_size[0], 1
    elif len(self.pre_size) == 2:
      height, width = self.pre_size
    else:
      raise ConnectorError(f'Currently, GridFour only supports the two-dimensional geometry.')

    conn_i = []
    conn_j = []
    for row in range(height):
      a = _grid_four(height, width, row, include_self=self.include_self)
      conn_i.extend(a[0])
      conn_j.extend(a[1])
    pre_ids = np.asarray(conn_i, dtype=IDX_DTYPE)
    post_ids = np.asarray(conn_j, dtype=IDX_DTYPE)

    return 'ij', (pre_ids, post_ids)

grid_four = GridFour()


# @tools.numba_jit
def _grid_n(height, width, row, n, include_self):
  conn_i = []
  conn_j = []
  for col in range(width):
    i_index = (row * width) + col
    for row_diff in range(-n, n + 1):
      for col_diff in range(-n, n + 1):
        if (not include_self) and (row_diff == col_diff == 0):
          continue
        if 0 <= row + row_diff < height and 0 <= col + col_diff < width:
          j_index = ((row + row_diff) * width) + col + col_diff
          conn_i.append(i_index)
          conn_j.append(j_index)
  return conn_i, conn_j


class GridN(OneEndConnector):
  """The nearest (2*N+1) * (2*N+1) neighbors conn method.

  Parameters
  ----------
  N : int
      Extend of the conn scope. For example:
      When N=1,
          [x x x]
          [x I x]
          [x x x]
      When N=2,
          [x x x x x]
          [x x x x x]
          [x x I x x]
          [x x x x x]
          [x x x x x]
  include_self : bool
      Whether create (i, i) conn ?
  """

  def __init__(self, N=1, include_self=False):
    super(GridN, self).__init__()
    self.N = N
    self.include_self = include_self

  def build_conn(self):
    if len(self.pre_size) == 1:
      height, width = self.pre_size[0], 1
    elif len(self.pre_size) == 2:
      height, width = self.pre_size
    else:
      raise ConnectorError(f'Currently, GridN only supports the two-dimensional geometry.')

    conn_i = []
    conn_j = []
    for row in range(height):
      res = _grid_n(height=height, width=width, row=row,
                    n=self.N, include_self=self.include_self)
      conn_i.extend(res[0])
      conn_j.extend(res[1])
    pre_ids = np.asarray(conn_i, dtype=IDX_DTYPE)
    post_ids = np.asarray(conn_j, dtype=IDX_DTYPE)

    return 'ij', (pre_ids, post_ids)


class GridEight(GridN):
  """The nearest eight neighbors conn method."""

  def __init__(self, include_self=False):
    super(GridEight, self).__init__(N=1, include_self=include_self)


grid_eight = GridEight()



# -*- coding: utf-8 -*-
import warnings

from brainpy.running import monitor

__all__ = [
  'Monitor'
]


class Monitor(monitor.Monitor):
  """Monitor class.

  .. deprecated:: 2.1.0
     Please use "brainpy.running.Monitor" instead.
  """
  def __init__(self, *args, **kwargs):
    warnings.warn('Please use "brainpy.running.Monitor" instead. '
                  '"brainpy.Monitor" is deprecated since version 2.1.0.',
                  DeprecationWarning)
    super(Monitor, self).__init__(*args, **kwargs)



# -*- coding: utf-8 -*-

from brainpy.base import naming

__all__ = [
  'DT',
  'F',

  'SUPPORTED_VAR_TYPE',
  'SCALAR_VAR',
  'POP_VAR',
  'SYSTEM_VAR',

  'SUPPORTED_WIENER_TYPE',
  'SCALAR_WIENER',
  'VECTOR_WIENER',

  'SUPPORTED_INTG_TYPE',
  'ITO_SDE',
  'STRA_SDE',

  'DE_INT',
  'ODE_INT',
  'SDE_INT',
  'DDE_INT',
  'FDE_INT',
  'PDE_INT',

  'unique_name',
]


DT = 'dt'
F = 'f'
G = 'g'




# Ito SDE_INT
# ---
# The SDE_INT integral proposed by Ito in 1940s.
ITO_SDE = 'Ito'

# Stratonovich SDE_INT
# ---
# The SDE_INT integral proposed by Stratonovich in 1960s.
STRA_SDE = 'Stratonovich'

SUPPORTED_INTG_TYPE = [
  ITO_SDE,
  STRA_SDE
]

# ------------------------------------------------------

# Scalar Wiener process
# ----
#
SCALAR_WIENER = 'scalar'

# Vector Wiener process
# ----
#
VECTOR_WIENER = 'vector'

SUPPORTED_WIENER_TYPE = [
  SCALAR_WIENER,
  VECTOR_WIENER
]

# ------------------------------------------------------

# Denotes each variable is a scalar variable
# -------
# For example:
#
# def derivative(a, b, t):
#     ...
#     return da, db
#
# The "a" and "b" are scalars: a=1, b=2
#
SCALAR_VAR = 'scalar'

# Denotes each variable is a homogeneous population
# -------
# For example:
#
# def derivative(a, b, t):
#     ...
#     return da, db
#
# The "a" and "b" are vectors or matrix:
#    a = np.array([1,2]),  b = np.array([3,4])
# or,
#    a = np.array([[1,2], [2,1]]),  b=np.array([[3,4], [4,3]])
#
POP_VAR = 'population'

# Denotes each variable is a system
# ------
# For example, the above defined differential equations can be defined as:
#
# def derivative(x, t):
#     a, b = x
#     ...
#     dx = np.array([da, db])
#     return dx
SYSTEM_VAR = 'system'

SUPPORTED_VAR_TYPE = [
  SCALAR_VAR,
  POP_VAR,
  SYSTEM_VAR,
]

# ------------------------------------------------------

# Differential equation type
# --------------------------

DE_INT = 'brainpy_itg_of'
ODE_INT = 'brainpy_itg_of_ode'
SDE_INT = 'brainpy_itg_of_sde'
DDE_INT = 'brainpy_itg_of_dde'
FDE_INT = 'brainpy_itg_of_fde'
PDE_INT = 'brainpy_itg_of_pde'


def unique_name(type):
  if type == 'ode':
    return naming.get_unique_name(ODE_INT)
  elif type == 'sde':
    return naming.get_unique_name(SDE_INT)
  elif type == 'dde':
    return naming.get_unique_name(DDE_INT)
  elif type == 'fde':
    return naming.get_unique_name(FDE_INT)
  elif type == 'pde':
    return naming.get_unique_name(PDE_INT)
  else:
    raise ValueError(f'Unknown differential equation type: {type}')



# -*- coding: utf-8 -*-

from typing import Union, Dict, Callable

import brainpy.math as bm
from brainpy.connect import TwoEndConnector
from brainpy.dyn.base import NeuGroup, TwoEndConn
from brainpy.initialize import Initializer
from brainpy.dyn.utils import init_delay
from brainpy.integrators import odeint, JointEq
from brainpy.types import Tensor, Parameter

__all__ = [
  'STP'
]


class STP(TwoEndConn):
  r"""Short-term plasticity model.

  **Model Descriptions**

  Short-term plasticity (STP) [1]_ [2]_ [3]_, also called dynamical synapses,
  refers to the changes of synaptic strengths over time in a way that reflects
  the history of presynaptic activity. Two types of STP, with opposite effects
  on synaptic efficacy, have been observed in experiments. They are known as
  Short-Term Depression (STD) and Short-Term Facilitation (STF).

  In the model proposed by Tsodyks and Markram [4]_ [5]_, the STD effect is
  modeled by a normalized variable :math:`x (0 \le x \le 1)`, denoting the fraction
  of resources that remain available after neurotransmitter depletion.
  The STF effect is modeled by a utilization parameter :math:`u`, representing
  the fraction of available resources ready for use (release probability).
  Following a spike,

  - (i) :math:`u` increases due to spike-induced calcium influx to the presynaptic
    terminal, after which
  - (ii) a fraction :math:`u` of available resources is consumed to produce the
    post-synaptic current.

  Between spikes, :math:`u` decays back to zero with time constant :math:`\tau_f`
  and :math:`x` recovers to 1 with time constant :math:`\tau_d`.

  In summary, the dynamics of STP is given by

  .. math::

      \begin{aligned}
      \frac{du}{dt} & =  -\frac{u}{\tau_f}+U(1-u^-)\delta(t-t_{sp}),\nonumber \\
      \frac{dx}{dt} & =  \frac{1-x}{\tau_d}-u^+x^-\delta(t-t_{sp}), \\
      \frac{dI}{dt} & =  -\frac{I}{\tau_s} + Au^+x^-\delta(t-t_{sp}),
      \end{aligned}

  where :math:`t_{sp}` denotes the spike time and :math:`U` is the increment
  of :math:`u` produced by a spike. :math:`u^-, x^-` are the corresponding
  variables just before the arrival of the spike, and :math:`u^+`
  refers to the moment just after the spike. The synaptic current generated
  at the synapse by the spike arriving at :math:`t_{sp}` is then given by

  .. math::

      \Delta I(t_{spike}) = Au^+x^-

  where :math:`A` denotes the response amplitude that would be produced
  by total release of all the neurotransmitter (:math:`u=x=1`), called
  absolute synaptic efficacy of the connections.

  **Model Examples**

  - `STP for Working Memory Capacity <https://brainpy-examples.readthedocs.io/en/latest/working_memory/Mi_2017_working_memory_capacity.html>`_

  **STD**

  .. plot::
    :include-source: True

    >>> import brainpy as bp
    >>> import matplotlib.pyplot as plt
    >>>
    >>> neu1 = bp.dyn.LIF(1)
    >>> neu2 = bp.dyn.LIF(1)
    >>> syn1 = bp.dyn.STP(neu1, neu2, bp.connect.All2All(), U=0.2, tau_d=150., tau_f=2.)
    >>> net = bp.dyn.Network(pre=neu1, syn=syn1, post=neu2)
    >>>
    >>> runner = bp.dyn.DSRunner(net, inputs=[('pre.input', 28.)], monitors=['syn.I', 'syn.u', 'syn.x'])
    >>> runner.run(150.)
    >>>
    >>>
    >>> # plot
    >>> fig, gs = bp.visualize.get_figure(2, 1, 3, 7)
    >>>
    >>> fig.add_subplot(gs[0, 0])
    >>> plt.plot(runner.mon.ts, runner.mon['syn.u'][:, 0], label='u')
    >>> plt.plot(runner.mon.ts, runner.mon['syn.x'][:, 0], label='x')
    >>> plt.legend()
    >>>
    >>> fig.add_subplot(gs[1, 0])
    >>> plt.plot(runner.mon.ts, runner.mon['syn.I'][:, 0], label='I')
    >>> plt.legend()
    >>>
    >>> plt.xlabel('Time (ms)')
    >>> plt.show()

  **STF**

  .. plot::
    :include-source: True

    >>> import brainpy as bp
    >>> import matplotlib.pyplot as plt
    >>>
    >>> neu1 = bp.dyn.LIF(1)
    >>> neu2 = bp.dyn.LIF(1)
    >>> syn1 = bp.dyn.STP(neu1, neu2, bp.connect.All2All(), U=0.1, tau_d=10, tau_f=100.)
    >>> net = bp.dyn.Network(pre=neu1, syn=syn1, post=neu2)
    >>>
    >>> runner = bp.dyn.DSRunner(net, inputs=[('pre.input', 28.)], monitors=['syn.I', 'syn.u', 'syn.x'])
    >>> runner.run(150.)
    >>>
    >>>
    >>> # plot
    >>> fig, gs = bp.visualize.get_figure(2, 1, 3, 7)
    >>>
    >>> fig.add_subplot(gs[0, 0])
    >>> plt.plot(runner.mon.ts, runner.mon['syn.u'][:, 0], label='u')
    >>> plt.plot(runner.mon.ts, runner.mon['syn.x'][:, 0], label='x')
    >>> plt.legend()
    >>>
    >>> fig.add_subplot(gs[1, 0])
    >>> plt.plot(runner.mon.ts, runner.mon['syn.I'][:, 0], label='I')
    >>> plt.legend()
    >>>
    >>> plt.xlabel('Time (ms)')
    >>> plt.show()



  **Model Parameters**

  ============= ============== ======== ===========================================
  **Parameter** **Init Value** **Unit** **Explanation**
  ------------- -------------- -------- -------------------------------------------
  tau_d         200            ms       Time constant of short-term depression.
  tau_f         1500           ms       Time constant of short-term facilitation.
  U             .15            \        The increment of :math:`u` produced by a spike.
  A             1              \        The response amplitude that would be produced by total release of all the neurotransmitter
  delay         0              ms       The decay time of the current :math:`I` output onto the post-synaptic neuron groups.
  ============= ============== ======== ===========================================


  **Model Variables**

  =============== ================== =====================================================================
  **Member name** **Initial values** **Explanation**
  --------------- ------------------ ---------------------------------------------------------------------
  u                 0                 Release probability of the neurotransmitters.
  x                 1                 A Normalized variable denoting the fraction of remain neurotransmitters.
  I                 0                 Synapse current output onto the post-synaptic neurons.
  =============== ================== =====================================================================

  **References**

  .. [1] Stevens, Charles F., and Yanyan Wang. "Facilitation and depression
         at single central synapses." Neuron 14, no. 4 (1995): 795-802.
  .. [2] Abbott, Larry F., J. A. Varela, Kamal Sen, and S. B. Nelson. "Synaptic
         depression and cortical gain control." Science 275, no. 5297 (1997): 221-224.
  .. [3] Abbott, L. F., and Wade G. Regehr. "Synaptic computation."
         Nature 431, no. 7010 (2004): 796-803.
  .. [4] Tsodyks, Misha, Klaus Pawelzik, and Henry Markram. "Neural networks
         with dynamic synapses." Neural computation 10.4 (1998): 821-835.
  .. [5] Tsodyks, Misha, and Si Wu. "Short-term synaptic plasticity."
         Scholarpedia 8, no. 10 (2013): 3153.

  """

  def __init__(
      self,
      pre: NeuGroup,
      post: NeuGroup,
      conn: Union[TwoEndConnector, Tensor, Dict[str, Tensor]],
      U: float = 0.15,
      tau_f: float = 1500.,
      tau_d: float = 200.,
      tau: float = 8.,
      A: float = 1.,
      delay_step: Union[int, Tensor, Initializer, Callable] = None,
      method: str = 'exp_auto',
      name: str = None
  ):
    super(STP, self).__init__(pre=pre, post=post, conn=conn, name=name)
    self.check_post_attrs('input')
    self.check_pre_attrs('spike')

    # parameters
    self.tau_d = tau_d
    self.tau_f = tau_f
    self.tau = tau
    self.U = U
    self.A = A

    # connections
    self.pre_ids, self.post_ids = self.conn.require('pre_ids', 'post_ids')

    # variables
    num = len(self.pre_ids)
    self.x = bm.Variable(bm.ones(num, dtype=bm.float_))
    self.u = bm.Variable(bm.zeros(num, dtype=bm.float_))
    self.I = bm.Variable(bm.zeros(num, dtype=bm.float_))
    self.delay_type, self.delay_step, self.delay_I = init_delay(delay_step, self.I)

    # integral
    self.integral = odeint(method=method, f=self.derivative)

  @property
  def derivative(self):
    dI = lambda I, t: -I / self.tau
    du = lambda u, t: - u / self.tau_f
    dx = lambda x, t: (1 - x) / self.tau_d
    return JointEq([dI, du, dx])

  def update(self, _t, _dt):
    # delayed pre-synaptic spikes
    if self.delay_type == 'homo':
      delayed_I = self.delay_I(self.delay_step)
    elif self.delay_type == 'heter':
      delayed_I = self.delay_I(self.delay_step, bm.arange(self.pre.num))
    else:
      delayed_I = self.I
    self.post.input += bm.syn2post(delayed_I, self.post_ids, self.post.num)
    self.I.value, u, x = self.integral(self.I, self.u, self.x, _t, dt=_dt)
    syn_sps = bm.pre2syn(self.pre.spike, self.pre_ids)
    u = bm.where(syn_sps, u + self.U * (1 - self.u), u)
    x = bm.where(syn_sps, x - u * self.x, x)
    self.I.value = bm.where(syn_sps, self.I + self.A * u * self.x, self.I)
    self.u.value = u
    self.x.value = x
    if self.delay_type in ['homo', 'heter']:
      self.delay_I.update(self.I)



# -*- coding: utf-8 -*-

import io
import os
import re

from setuptools import find_packages
from setuptools import setup

# --------------For pip install backup plan--------------

# from pip._internal.utils.compat import stdlib_pkgs
# from typing import cast
# def get_installed_distributions(
#     local_only: bool = True,
#     skip = stdlib_pkgs,
#     include_editables: bool = True,
#     editables_only: bool = False,
#     user_only: bool = False,
#     paths = None,
# ):
#   """Return a list of installed Distribution objects.
#   Left for compatibility until direct pkg_resources uses are refactored out.
#   """
#   from pip._internal.metadata import get_default_environment, get_environment
#   from pip._internal.metadata.pkg_resources import Distribution as _Dist
#
#   if paths is None:
#     env = get_default_environment()
#   else:
#     env = get_environment(paths)
#   dists = env.iter_installed_distributions(
#     local_only=local_only,
#     skip=skip,
#     include_editables=include_editables,
#     editables_only=editables_only,
#     user_only=user_only,
#   )
#   return [cast(_Dist, dist)._dist for dist in dists]

# ----------------------------------------------------

try:
  import pkg_resources
  installed_packages = pkg_resources.working_set
  for i in installed_packages:
    if i.key == 'brainpy-simulator':
      raise SystemError('Please uninstall the older version of brainpy '
                        f'package "brainpy-simulator={i.version}" '
                        f'(located in {i.location}) first. \n'
                        '>>> pip uninstall brainpy-simulator')
    if i.key == 'brain-py':
      raise SystemError('Please uninstall the older version of brainpy '
                        f'package "brain-py={i.version}" '
                        f'(located in {i.location}) first. \n'
                        '>>> pip uninstall brain-py')
except ModuleNotFoundError:
  pass


# version
here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, 'brainpy', '__init__.py'), 'r') as f:
  init_py = f.read()
version = re.search('__version__ = "(.*)"', init_py).groups()[0]

# obtain long description from README
with io.open(os.path.join(here, 'README.md'), 'r', encoding='utf-8') as f:
  README = f.read()

# setup
setup(
  name='brainpy',
  version=version,
  description='BrainPy: Brain Dynamics Programming in Python',
  long_description=README,
  long_description_content_type="text/markdown",
  author='BrainPy Team',
  author_email='chao.brain@qq.com',
  packages=find_packages(),
  python_requires='>=3.7',
  install_requires=[
    'numpy>=1.15',
    'jax>=0.3.0',
    'tqdm',
  ],
  extras_require={
    'cpu': ['jaxlib>=0.3.0', 'brainpylib>=0.0.4'],
    'cuda': ['jaxlib>=0.3.0', 'brainpylib>=0.0.4'],
    'all': ['jaxlib>=0.3.0', 'brainpylib>=0.0.4',
            'numba>=0.50', 'scipy>=1.1.0',
            'networkx', 'matplotlib']
  },
  url='https://github.com/PKU-NIP-Lab/BrainPy',
  keywords='computational neuroscience, '
           'brain-inspired computation, '
           'dynamical systems, '
           'differential equations, '
           'brain modeling, '
           'brain dynamics programming',
  classifiers=[
    'Natural Language :: English',
    'Operating System :: OS Independent',
    'Programming Language :: Python',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
    'Programming Language :: Python :: 3.10',
    'Intended Audience :: Science/Research',
    'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
    'Topic :: Scientific/Engineering :: Bio-Informatics',
    'Topic :: Scientific/Engineering :: Mathematics',
    'Topic :: Scientific/Engineering :: Artificial Intelligence',
    'Topic :: Software Development :: Libraries',
  ],
  license='GPL-3.0 License',
)



# -*- coding: utf-8 -*-

from typing import Union, Callable

import numpy as np
from jax.experimental.host_callback import id_tap

import brainpy.math as bm
from brainpy import check
from brainpy.dyn.base import NeuGroup
from brainpy.initialize import Initializer, Uniform, init_param
from brainpy.integrators.dde import ddeint
from brainpy.integrators.joint_eq import JointEq
from brainpy.integrators.ode import odeint
from brainpy.tools.checking import check_float, check_initializer
from brainpy.types import Shape, Tensor
from .noise_models import OUProcess

__all__ = [
  'RateGroup',
  'RateFHN',
  'FeedbackFHN',
  'RateQIF',
  'StuartLandauOscillator',
  'WilsonCowanModel',
]


class RateGroup(NeuGroup):
  def update(self, _t, _dt):
    raise NotImplementedError


class RateFHN(NeuGroup):
  r"""FitzHugh-Nagumo system used in [1]_.

  .. math::

     \frac{dx}{dt} = -\alpha V^3 + \beta V^2 + \gamma V - w + I_{ext}\\
     \tau \frac{dy}{dt} = (V - \delta  - \epsilon w)

  Parameters
  ----------
  size: Shape
    The model size.
  x_ou_mean: Parameter
    The noise mean of the :math:`x` variable, [mV/ms]
  y_ou_mean: Parameter
    The noise mean of the :math:`y` variable, [mV/ms].
  x_ou_sigma: Parameter
    The noise intensity of the :math:`x` variable, [mV/ms/sqrt(ms)].
  y_ou_sigma: Parameter
    The noise intensity of the :math:`y` variable, [mV/ms/sqrt(ms)].
  x_ou_tau: Parameter
    The timescale of the Ornstein-Uhlenbeck noise process of :math:`x` variable, [ms].
  y_ou_tau: Parameter
    The timescale of the Ornstein-Uhlenbeck noise process of :math:`y` variable, [ms].


  References
  ----------
  .. [1] Kostova, T., Ravindran, R., & Schonbek, M. (2004). FitzHugh–Nagumo
         revisited: Types of bifurcations, periodical forcing and stability
         regions by a Lyapunov functional. International journal of
         bifurcation and chaos, 14(03), 913-925.

  """

  def __init__(
      self,
      size: Shape,

      # fhn parameters
      alpha: Union[float, Tensor, Initializer, Callable] = 3.0,
      beta: Union[float, Tensor, Initializer, Callable] = 4.0,
      gamma: Union[float, Tensor, Initializer, Callable] = -1.5,
      delta: Union[float, Tensor, Initializer, Callable] = 0.0,
      epsilon: Union[float, Tensor, Initializer, Callable] = 0.5,
      tau: Union[float, Tensor, Initializer, Callable] = 20.0,

      # noise parameters
      x_ou_mean: Union[float, Tensor, Initializer, Callable] = 0.0,
      x_ou_sigma: Union[float, Tensor, Initializer, Callable] = 0.0,
      x_ou_tau: Union[float, Tensor, Initializer, Callable] = 5.0,
      y_ou_mean: Union[float, Tensor, Initializer, Callable] = 0.0,
      y_ou_sigma: Union[float, Tensor, Initializer, Callable] = 0.0,
      y_ou_tau: Union[float, Tensor, Initializer, Callable] = 5.0,

      # other parameters
      x_initializer: Union[Initializer, Callable, Tensor] = Uniform(0, 0.05),
      y_initializer: Union[Initializer, Callable, Tensor] = Uniform(0, 0.05),
      method: str = None,
      sde_method: str = None,
      name: str = None,
  ):
    super(RateFHN, self).__init__(size=size, name=name)

    # model parameters
    self.alpha = init_param(alpha, self.num, allow_none=False)
    self.beta = init_param(beta, self.num, allow_none=False)
    self.gamma = init_param(gamma, self.num, allow_none=False)
    self.delta = init_param(delta, self.num, allow_none=False)
    self.epsilon = init_param(epsilon, self.num, allow_none=False)
    self.tau = init_param(tau, self.num, allow_none=False)

    # noise parameters
    self.x_ou_mean = init_param(x_ou_mean, self.num, allow_none=False)  # mV/ms, OU process
    self.y_ou_mean = init_param(y_ou_mean, self.num, allow_none=False)  # mV/ms, OU process
    self.x_ou_sigma = init_param(x_ou_sigma, self.num, allow_none=False)  # mV/ms/sqrt(ms), noise intensity
    self.y_ou_sigma = init_param(y_ou_sigma, self.num, allow_none=False)  # mV/ms/sqrt(ms), noise intensity
    self.x_ou_tau = init_param(x_ou_tau, self.num,
                               allow_none=False)  # ms, timescale of the Ornstein-Uhlenbeck noise process
    self.y_ou_tau = init_param(y_ou_tau, self.num,
                               allow_none=False)  # ms, timescale of the Ornstein-Uhlenbeck noise process

    # variables
    check_initializer(x_initializer, 'x_initializer')
    check_initializer(y_initializer, 'y_initializer')
    self.x = bm.Variable(init_param(x_initializer, (self.num,)))
    self.y = bm.Variable(init_param(x_initializer, (self.num,)))
    self.input = bm.Variable(bm.zeros(self.num))

    # noise variables
    self.x_ou = self.y_ou = None
    if bm.any(self.x_ou_mean > 0.) or bm.any(self.x_ou_sigma > 0.):
      self.x_ou = OUProcess(self.num,
                            self.x_ou_mean, self.x_ou_sigma, self.x_ou_tau,
                            method=sde_method)
    if bm.any(self.y_ou_mean > 0.) or bm.any(self.y_ou_sigma > 0.):
      self.y_ou = OUProcess(self.num,
                            self.y_ou_mean, self.y_ou_sigma, self.y_ou_tau,
                            method=sde_method)

    # integral functions
    self.integral = odeint(f=JointEq([self.dx, self.dy]), method=method)

  def dx(self, x, t, y, x_ext):
    return - self.alpha * x ** 3 + self.beta * x ** 2 + self.gamma * x - y + x_ext

  def dy(self, y, t, x, y_ext=0.):
    return (x - self.delta - self.epsilon * y) / self.tau + y_ext

  def update(self, _t, _dt):
    if self.x_ou is not None:
      self.input += self.x_ou.x
      self.x_ou.update(_t, _dt)
    y_ext = 0.
    if self.y_ou is not None:
      y_ext = self.y_ou.x
      self.y_ou.update(_t, _dt)
    x, y = self.integral(self.x, self.y, _t, x_ext=self.input, y_ext=y_ext, dt=_dt)
    self.x.value = x
    self.y.value = y
    self.input[:] = 0.


class FeedbackFHN(NeuGroup):
  r"""FitzHugh-Nagumo model with recurrent neural feedback.

  The equation of the feedback FitzHugh-Nagumo model [4]_ is given by

  .. math::

     \begin{aligned}
     \frac{dx}{dt} &= x(t) - \frac{x^3(t)}{3} - y(t) + \mu[x(t-\mathrm{delay}) - x_0] \\
     \frac{dy}{dt} &= [x(t) + a - b y(t)] / \tau
     \end{aligned}


  **Model Examples**

  >>> import brainpy as bp
  >>> fhn = bp.dyn.FeedbackFHN(1, delay=10.)
  >>> runner = bp.dyn.DSRunner(fhn, inputs=('input', 1.), monitors=['x', 'y'])
  >>> runner.run(100.)
  >>> bp.visualize.line_plot(runner.mon.ts, runner.mon.y, legend='y')
  >>> bp.visualize.line_plot(runner.mon.ts, runner.mon.x, legend='x', show=True)


  **Model Parameters**

  ============= ============== ======== ========================
  **Parameter** **Init Value** **Unit** **Explanation**
  ------------- -------------- -------- ------------------------
  a             1              \        Positive constant
  b             1              \        Positive constant
  tau           12.5           ms       Membrane time constant.
  delay         10             ms       Synaptic delay time constant.
  V_th          1.8            mV       Threshold potential of spike.
  v0            -1             mV       Resting potential.
  mu            1.8            \        The feedback strength. When positive, it is a excitatory feedback;
                                        when negative, it is a inhibitory feedback.
  ============= ============== ======== ========================

  Parameters
  ----------
  x_ou_mean: Parameter
    The noise mean of the :math:`x` variable, [mV/ms]
  y_ou_mean: Parameter
    The noise mean of the :math:`y` variable, [mV/ms].
  x_ou_sigma: Parameter
    The noise intensity of the :math:`x` variable, [mV/ms/sqrt(ms)].
  y_ou_sigma: Parameter
    The noise intensity of the :math:`y` variable, [mV/ms/sqrt(ms)].
  x_ou_tau: Parameter
    The timescale of the Ornstein-Uhlenbeck noise process of :math:`x` variable, [ms].
  y_ou_tau: Parameter
    The timescale of the Ornstein-Uhlenbeck noise process of :math:`y` variable, [ms].



  References
  ----------
  .. [4] Plant, Richard E. (1981). *A FitzHugh Differential-Difference
         Equation Modeling Recurrent Neural Feedback. SIAM Journal on
         Applied Mathematics, 40(1), 150–162.* doi:10.1137/0140012

  """

  def __init__(
      self,
      size: Shape,

      # model parameters
      a: Union[float, Tensor, Initializer, Callable] = 0.7,
      b: Union[float, Tensor, Initializer, Callable] = 0.8,
      delay: Union[float, Tensor, Initializer, Callable] = 10.,
      tau: Union[float, Tensor, Initializer, Callable] = 12.5,
      mu: Union[float, Tensor, Initializer, Callable] = 1.6886,
      v0: Union[float, Tensor, Initializer, Callable] = -1,

      # noise parameters
      x_ou_mean: Union[float, Tensor, Initializer, Callable] = 0.0,
      x_ou_sigma: Union[float, Tensor, Initializer, Callable] = 0.0,
      x_ou_tau: Union[float, Tensor, Initializer, Callable] = 5.0,
      y_ou_mean: Union[float, Tensor, Initializer, Callable] = 0.0,
      y_ou_sigma: Union[float, Tensor, Initializer, Callable] = 0.0,
      y_ou_tau: Union[float, Tensor, Initializer, Callable] = 5.0,

      # other parameters
      x_initializer: Union[Initializer, Callable, Tensor] = Uniform(0, 0.05),
      y_initializer: Union[Initializer, Callable, Tensor] = Uniform(0, 0.05),
      method: str = 'rk4',
      sde_method: str = None,
      name: str = None,
      dt: float = None
  ):
    super(FeedbackFHN, self).__init__(size=size, name=name)

    # dt
    self.dt = bm.get_dt() if dt is None else dt
    check_float(self.dt, 'dt', allow_none=False, min_bound=0., allow_int=False)

    # parameters
    self.a = init_param(a, self.num, allow_none=False)
    self.b = init_param(b, self.num, allow_none=False)
    self.delay = init_param(delay, self.num, allow_none=False)
    self.tau = init_param(tau, self.num, allow_none=False)
    self.mu = init_param(mu, self.num, allow_none=False)  # feedback strength
    self.v0 = init_param(v0, self.num, allow_none=False)  # resting potential

    # noise parameters
    self.x_ou_mean = init_param(x_ou_mean, self.num, allow_none=False)
    self.y_ou_mean = init_param(y_ou_mean, self.num, allow_none=False)
    self.x_ou_sigma = init_param(x_ou_sigma, self.num, allow_none=False)
    self.y_ou_sigma = init_param(y_ou_sigma, self.num, allow_none=False)
    self.x_ou_tau = init_param(x_ou_tau, self.num, allow_none=False)
    self.y_ou_tau = init_param(y_ou_tau, self.num, allow_none=False)

    # variables
    check_initializer(x_initializer, 'x_initializer')
    check_initializer(y_initializer, 'y_initializer')
    self.x = bm.Variable(init_param(x_initializer, (self.num,)))
    self.y = bm.Variable(init_param(x_initializer, (self.num,)))
    self.x_delay = bm.TimeDelay(self.x, self.delay, dt=self.dt, interp_method='round')
    self.input = bm.Variable(bm.zeros(self.num))

    # noise variables
    self.x_ou = self.y_ou = None
    if bm.any(self.x_ou_mean > 0.) or bm.any(self.x_ou_sigma > 0.):
      self.x_ou = OUProcess(self.num,
                            self.x_ou_mean, self.x_ou_sigma, self.x_ou_tau,
                            method=sde_method)
    if bm.any(self.y_ou_mean > 0.) or bm.any(self.y_ou_sigma > 0.):
      self.y_ou = OUProcess(self.num,
                            self.y_ou_mean, self.y_ou_sigma, self.y_ou_tau,
                            method=sde_method)

    # integral
    self.integral = ddeint(method=method,
                           f=JointEq([self.dx, self.dy]),
                           state_delays={'V': self.x_delay})

  def dx(self, x, t, y, x_ext):
    return x - x * x * x / 3 - y + x_ext + self.mu * (self.x_delay(t - self.delay) - self.v0)

  def dy(self, y, t, x, y_ext):
    return (x + self.a - self.b * y + y_ext) / self.tau

  def _check_dt(self, dt, *args):
    if np.absolute(dt - self.dt) > 1e-6:
      raise ValueError(f'The "dt" {dt} used in model running is '
                       f'not consistent with the "dt" {self.dt} '
                       f'used in model definition.')

  def update(self, _t, _dt):
    if check.is_checking():
      id_tap(self._check_dt, _dt)
    if self.x_ou is not None:
      self.input += self.x_ou.x
      self.x_ou.update(_t, _dt)
    y_ext = 0.
    if self.y_ou is not None:
      y_ext = self.y_ou.x
      self.y_ou.update(_t, _dt)
    x, y = self.integral(self.x, self.y, _t, x_ext=self.input, y_ext=y_ext, dt=_dt)
    self.x.value = x
    self.y.value = y
    self.input[:] = 0.


class RateQIF(NeuGroup):
  r"""A mean-field model of a quadratic integrate-and-fire neuron population.

  **Model Descriptions**

  The QIF population mean-field model, which has been derived from a
  population of all-to-all coupled QIF neurons in [5]_.
  The model equations are given by:

  .. math::

     \begin{aligned}
     \tau \dot{r} &=\frac{\Delta}{\pi \tau}+2 r v \\
     \tau \dot{v} &=v^{2}+\bar{\eta}+I(t)+J r \tau-(\pi r \tau)^{2}
     \end{aligned}

  where :math:`r` is the average firing rate and :math:`v` is the
  average membrane potential of the QIF population [5]_.

  This mean-field model is an exact representation of the macroscopic
  firing rate and membrane potential dynamics of a spiking neural network
  consisting of QIF neurons with Lorentzian distributed background
  excitabilities. While the mean-field derivation is mathematically
  only valid for all-to-all coupled populations of infinite size, it
  has been shown that there is a close correspondence between the
  mean-field model and neural populations with sparse coupling and
  population sizes of a few thousand neurons [6]_.

  **Model Parameters**

  ============= ============== ======== ========================
  **Parameter** **Init Value** **Unit** **Explanation**
  ------------- -------------- -------- ------------------------
  tau           1              ms       the population time constant
  eta           -5.            \        the mean of a Lorenzian distribution over the neural excitability in the population
  delta         1.0            \        the half-width at half maximum of the Lorenzian distribution over the neural excitability
  J             15             \        the strength of the recurrent coupling inside the population
  ============= ============== ======== ========================

  Parameters
  ----------
  x_ou_mean: Parameter
    The noise mean of the :math:`x` variable, [mV/ms]
  y_ou_mean: Parameter
    The noise mean of the :math:`y` variable, [mV/ms].
  x_ou_sigma: Parameter
    The noise intensity of the :math:`x` variable, [mV/ms/sqrt(ms)].
  y_ou_sigma: Parameter
    The noise intensity of the :math:`y` variable, [mV/ms/sqrt(ms)].
  x_ou_tau: Parameter
    The timescale of the Ornstein-Uhlenbeck noise process of :math:`x` variable, [ms].
  y_ou_tau: Parameter
    The timescale of the Ornstein-Uhlenbeck noise process of :math:`y` variable, [ms].


  References
  ----------
  .. [5] E. Montbrió, D. Pazó, A. Roxin (2015) Macroscopic description for
         networks of spiking neurons. Physical Review X, 5:021028,
         https://doi.org/10.1103/PhysRevX.5.021028.
  .. [6] R. Gast, H. Schmidt, T.R. Knösche (2020) A Mean-Field Description
         of Bursting Dynamics in Spiking Neural Networks with Short-Term
         Adaptation. Neural Computation 32.9 (2020): 1615-1634.

  """

  def __init__(
      self,
      size: Shape,

      # model parameters
      tau: Union[float, Tensor, Initializer, Callable] = 1.,
      eta: Union[float, Tensor, Initializer, Callable] = -5.0,
      delta: Union[float, Tensor, Initializer, Callable] = 1.0,
      J: Union[float, Tensor, Initializer, Callable] = 15.,

      # noise parameters
      x_ou_mean: Union[float, Tensor, Initializer, Callable] = 0.0,
      x_ou_sigma: Union[float, Tensor, Initializer, Callable] = 0.0,
      x_ou_tau: Union[float, Tensor, Initializer, Callable] = 5.0,
      y_ou_mean: Union[float, Tensor, Initializer, Callable] = 0.0,
      y_ou_sigma: Union[float, Tensor, Initializer, Callable] = 0.0,
      y_ou_tau: Union[float, Tensor, Initializer, Callable] = 5.0,

      # other parameters
      x_initializer: Union[Initializer, Callable, Tensor] = Uniform(0, 0.05),
      y_initializer: Union[Initializer, Callable, Tensor] = Uniform(0, 0.05),
      method: str = 'exp_auto',
      name: str = None,
      sde_method: str = None,
  ):
    super(RateQIF, self).__init__(size=size, name=name)

    # parameters
    self.tau = init_param(tau, self.num, allow_none=False)
    # the mean of a Lorenzian distribution over the neural excitability in the population
    self.eta = init_param(eta, self.num, allow_none=False)
    # the half-width at half maximum of the Lorenzian distribution over the neural excitability
    self.delta = init_param(delta, self.num, allow_none=False)
    # the strength of the recurrent coupling inside the population
    self.J = init_param(J, self.num, allow_none=False)

    # noise parameters
    self.x_ou_mean = init_param(x_ou_mean, self.num, allow_none=False)
    self.y_ou_mean = init_param(y_ou_mean, self.num, allow_none=False)
    self.x_ou_sigma = init_param(x_ou_sigma, self.num, allow_none=False)
    self.y_ou_sigma = init_param(y_ou_sigma, self.num, allow_none=False)
    self.x_ou_tau = init_param(x_ou_tau, self.num, allow_none=False)
    self.y_ou_tau = init_param(y_ou_tau, self.num, allow_none=False)

    # variables
    check_initializer(x_initializer, 'x_initializer')
    check_initializer(y_initializer, 'y_initializer')
    self.x = bm.Variable(init_param(x_initializer, (self.num,)))
    self.y = bm.Variable(init_param(x_initializer, (self.num,)))
    self.input = bm.Variable(bm.zeros(self.num))

    # noise variables
    self.x_ou = self.y_ou = None
    if bm.any(self.x_ou_mean > 0.) or bm.any(self.x_ou_sigma > 0.):
      self.x_ou = OUProcess(self.num,
                            self.x_ou_mean, self.x_ou_sigma, self.x_ou_tau,
                            method=sde_method)
    if bm.any(self.y_ou_mean > 0.) or bm.any(self.y_ou_sigma > 0.):
      self.y_ou = OUProcess(self.num,
                            self.y_ou_mean, self.y_ou_sigma, self.y_ou_tau,
                            method=sde_method)

    # functions
    self.integral = odeint(JointEq([self.dx, self.dy]), method=method)

  def dy(self, y, t, x, y_ext):
    return (self.delta / (bm.pi * self.tau) + 2. * x * y + y_ext) / self.tau

  def dx(self, x, t, y, x_ext):
    return (x ** 2 + self.eta + x_ext + self.J * y * self.tau -
            (bm.pi * y * self.tau) ** 2) / self.tau

  def update(self, _t, _dt):
    if self.x_ou is not None:
      self.input += self.x_ou.x
      self.x_ou.update(_t, _dt)
    y_ext = 0.
    if self.y_ou is not None:
      y_ext = self.y_ou.x
      self.y_ou.update(_t, _dt)
    x, y = self.integral(self.x, self.y, t=_t, x_ext=self.input, y_ext=y_ext, dt=_dt)
    self.x.value = x
    self.y.value = y
    self.input[:] = 0.


class StuartLandauOscillator(RateGroup):
  r"""
  Stuart-Landau model with Hopf bifurcation.

  .. math::

     \frac{dx}{dt} = (a - x^2 - y^2) * x - w*y + I^x_{ext} \\
     \frac{dy}{dt} = (a - x^2 - y^2) * y + w*x + I^y_{ext}

  Parameters
  ----------
  x_ou_mean: Parameter
    The noise mean of the :math:`x` variable, [mV/ms]
  y_ou_mean: Parameter
    The noise mean of the :math:`y` variable, [mV/ms].
  x_ou_sigma: Parameter
    The noise intensity of the :math:`x` variable, [mV/ms/sqrt(ms)].
  y_ou_sigma: Parameter
    The noise intensity of the :math:`y` variable, [mV/ms/sqrt(ms)].
  x_ou_tau: Parameter
    The timescale of the Ornstein-Uhlenbeck noise process of :math:`x` variable, [ms].
  y_ou_tau: Parameter
    The timescale of the Ornstein-Uhlenbeck noise process of :math:`y` variable, [ms].

  """

  def __init__(
      self,
      size: Shape,

      # model parameters
      a: Union[float, Tensor, Initializer, Callable] = 0.25,
      w: Union[float, Tensor, Initializer, Callable] = 0.2,

      # noise parameters
      x_ou_mean: Union[float, Tensor, Initializer, Callable] = 0.0,
      x_ou_sigma: Union[float, Tensor, Initializer, Callable] = 0.0,
      x_ou_tau: Union[float, Tensor, Initializer, Callable] = 5.0,
      y_ou_mean: Union[float, Tensor, Initializer, Callable] = 0.0,
      y_ou_sigma: Union[float, Tensor, Initializer, Callable] = 0.0,
      y_ou_tau: Union[float, Tensor, Initializer, Callable] = 5.0,

      # other parameters
      x_initializer: Union[Initializer, Callable, Tensor] = Uniform(0, 0.5),
      y_initializer: Union[Initializer, Callable, Tensor] = Uniform(0, 0.5),
      method: str = None,
      sde_method: str = None,
      name: str = None,
  ):
    super(StuartLandauOscillator, self).__init__(size=size,
                                                 name=name)

    # model parameters
    self.a = init_param(a, self.num, allow_none=False)
    self.w = init_param(w, self.num, allow_none=False)

    # noise parameters
    self.x_ou_mean = init_param(x_ou_mean, self.num, allow_none=False)
    self.y_ou_mean = init_param(y_ou_mean, self.num, allow_none=False)
    self.x_ou_sigma = init_param(x_ou_sigma, self.num, allow_none=False)
    self.y_ou_sigma = init_param(y_ou_sigma, self.num, allow_none=False)
    self.x_ou_tau = init_param(x_ou_tau, self.num, allow_none=False)
    self.y_ou_tau = init_param(y_ou_tau, self.num, allow_none=False)

    # variables
    check_initializer(x_initializer, 'x_initializer')
    check_initializer(y_initializer, 'y_initializer')
    self.x = bm.Variable(init_param(x_initializer, (self.num,)))
    self.y = bm.Variable(init_param(x_initializer, (self.num,)))
    self.input = bm.Variable(bm.zeros(self.num))

    # noise variables
    self.x_ou = self.y_ou = None
    if bm.any(self.x_ou_mean > 0.) or bm.any(self.x_ou_sigma > 0.):
      self.x_ou = OUProcess(self.num,
                            self.x_ou_mean, self.x_ou_sigma, self.x_ou_tau,
                            method=sde_method)
    if bm.any(self.y_ou_mean > 0.) or bm.any(self.y_ou_sigma > 0.):
      self.y_ou = OUProcess(self.num,
                            self.y_ou_mean, self.y_ou_sigma, self.y_ou_tau,
                            method=sde_method)

    # integral functions
    self.integral = odeint(f=JointEq([self.dx, self.dy]), method=method)

  def dx(self, x, t, y, x_ext, a, w):
    return (a - x * x - y * y) * x - w * y + x_ext

  def dy(self, y, t, x, y_ext, a, w):
    return (a - x * x - y * y) * y - w * y + y_ext

  def update(self, _t, _dt):
    if self.x_ou is not None:
      self.input += self.x_ou.x
      self.x_ou.update(_t, _dt)
    y_ext = 0.
    if self.y_ou is not None:
      y_ext = self.y_ou.x
      self.y_ou.update(_t, _dt)
    x, y = self.integral(self.x, self.y, _t, x_ext=self.input,
                         y_ext=y_ext, a=self.a, w=self.w, dt=_dt)
    self.x.value = x
    self.y.value = y
    self.input[:] = 0.


class WilsonCowanModel(RateGroup):
  """Wilson-Cowan population model.


  Parameters
  ----------
  x_ou_mean: Parameter
    The noise mean of the :math:`x` variable, [mV/ms]
  y_ou_mean: Parameter
    The noise mean of the :math:`y` variable, [mV/ms].
  x_ou_sigma: Parameter
    The noise intensity of the :math:`x` variable, [mV/ms/sqrt(ms)].
  y_ou_sigma: Parameter
    The noise intensity of the :math:`y` variable, [mV/ms/sqrt(ms)].
  x_ou_tau: Parameter
    The timescale of the Ornstein-Uhlenbeck noise process of :math:`x` variable, [ms].
  y_ou_tau: Parameter
    The timescale of the Ornstein-Uhlenbeck noise process of :math:`y` variable, [ms].


  """

  def __init__(
      self,
      size: Shape,

      # Excitatory parameters
      E_tau: Union[float, Tensor, Initializer, Callable] = 1.,  # excitatory time constant
      E_a: Union[float, Tensor, Initializer, Callable] = 1.2,  # excitatory gain
      E_theta: Union[float, Tensor, Initializer, Callable] = 2.8,  # excitatory firing threshold

      # Inhibitory parameters
      I_tau: Union[float, Tensor, Initializer, Callable] = 1.,  # inhibitory time constant
      I_a: Union[float, Tensor, Initializer, Callable] = 1.,  # inhibitory gain
      I_theta: Union[float, Tensor, Initializer, Callable] = 4.0,  # inhibitory firing threshold

      # connection parameters
      wEE: Union[float, Tensor, Initializer, Callable] = 12.,  # local E-E coupling
      wIE: Union[float, Tensor, Initializer, Callable] = 4.,  # local E-I coupling
      wEI: Union[float, Tensor, Initializer, Callable] = 13.,  # local I-E coupling
      wII: Union[float, Tensor, Initializer, Callable] = 11.,  # local I-I coupling

      # Refractory parameter
      r: Union[float, Tensor, Initializer, Callable] = 1,

      # noise parameters
      x_ou_mean: Union[float, Tensor, Initializer, Callable] = 0.0,
      x_ou_sigma: Union[float, Tensor, Initializer, Callable] = 0.0,
      x_ou_tau: Union[float, Tensor, Initializer, Callable] = 5.0,
      y_ou_mean: Union[float, Tensor, Initializer, Callable] = 0.0,
      y_ou_sigma: Union[float, Tensor, Initializer, Callable] = 0.0,
      y_ou_tau: Union[float, Tensor, Initializer, Callable] = 5.0,

      # state initializer
      x_initializer: Union[Initializer, Callable, Tensor] = Uniform(max_val=0.05),
      y_initializer: Union[Initializer, Callable, Tensor] = Uniform(max_val=0.05),

      # other parameters
      sde_method: str = None,
      method: str = 'exp_euler_auto',
      name: str = None,
  ):
    super(WilsonCowanModel, self).__init__(size=size, name=name)

    # model parameters
    self.E_a = init_param(E_a, self.num, allow_none=False)
    self.I_a = init_param(I_a, self.num, allow_none=False)
    self.E_tau = init_param(E_tau, self.num, allow_none=False)
    self.I_tau = init_param(I_tau, self.num, allow_none=False)
    self.E_theta = init_param(E_theta, self.num, allow_none=False)
    self.I_theta = init_param(I_theta, self.num, allow_none=False)
    self.wEE = init_param(wEE, self.num, allow_none=False)
    self.wIE = init_param(wIE, self.num, allow_none=False)
    self.wEI = init_param(wEI, self.num, allow_none=False)
    self.wII = init_param(wII, self.num, allow_none=False)
    self.r = init_param(r, self.num, allow_none=False)

    # noise parameters
    self.x_ou_mean = init_param(x_ou_mean, self.num, allow_none=False)
    self.y_ou_mean = init_param(y_ou_mean, self.num, allow_none=False)
    self.x_ou_sigma = init_param(x_ou_sigma, self.num, allow_none=False)
    self.y_ou_sigma = init_param(y_ou_sigma, self.num, allow_none=False)
    self.x_ou_tau = init_param(x_ou_tau, self.num, allow_none=False)
    self.y_ou_tau = init_param(y_ou_tau, self.num, allow_none=False)

    # variables
    check_initializer(x_initializer, 'x_initializer')
    check_initializer(y_initializer, 'y_initializer')
    self.x = bm.Variable(init_param(x_initializer, (self.num,)))
    self.y = bm.Variable(init_param(x_initializer, (self.num,)))
    self.input = bm.Variable(bm.zeros(self.num))

    # noise variables
    self.x_ou = self.y_ou = None
    if bm.any(self.x_ou_mean > 0.) or bm.any(self.x_ou_sigma > 0.):
      self.x_ou = OUProcess(self.num,
                            self.x_ou_mean, self.x_ou_sigma, self.x_ou_tau,
                            method=sde_method)
    if bm.any(self.y_ou_mean > 0.) or bm.any(self.y_ou_sigma > 0.):
      self.y_ou = OUProcess(self.num,
                            self.y_ou_mean, self.y_ou_sigma, self.y_ou_tau,
                            method=sde_method)

    # functions
    self.integral = odeint(f=JointEq([self.dx, self.dy]), method=method)

  # functions
  def F(self, x, a, theta):
    return 1 / (1 + bm.exp(-a * (x - theta))) - 1 / (1 + bm.exp(a * theta))

  def dx(self, x, t, y, x_ext):
    x = self.wEE * x - self.wIE * y + x_ext
    return (-x + (1 - self.r * x) * self.F(x, self.E_a, self.E_theta)) / self.E_tau

  def dy(self, y, t, x, y_ext):
    x = self.wEI * x - self.wII * y + y_ext
    return (-y + (1 - self.r * y) * self.F(x, self.I_a, self.I_theta)) / self.I_tau

  def update(self, _t, _dt):
    if self.x_ou is not None:
      self.input += self.x_ou.x
      self.x_ou.update(_t, _dt)
    y_ext = 0.
    if self.y_ou is not None:
      y_ext = self.y_ou.x
      self.y_ou.update(_t, _dt)
    x, y = self.integral(self.x, self.y, _t, x_ext=self.input, y_ext=y_ext, dt=_dt)
    self.x.value = x
    self.y.value = y
    self.input[:] = 0.


class JansenRitModel(RateGroup):
  pass


class KuramotoOscillator(RateGroup):
  pass


class ThetaNeuron(RateGroup):
  pass


class RateQIFWithSFA(RateGroup):
  pass


class VanDerPolOscillator(RateGroup):
  pass



# -*- coding: utf-8 -*-


import numpy as np
from jax import numpy as jnp
from jax.tree_util import register_pytree_node

from brainpy.errors import MathError

__all__ = [
  'JaxArray',
  'ndarray',  # alias of JaxArray
  'Variable',
  'TrainVar',
  'Parameter',
]

# Ways to change values in a zero-dimensional array
# -----
# Reference: https://stackoverflow.com/questions/56954714/how-do-i-assign-to-a-zero-dimensional-numpy-array
#
#   >>> x = np.array(10)
# 1. index the original array with ellipsis or an empty tuple
#    >>> x[...] = 2
#    >>> x[()] = 2

_all_slice = slice(None, None, None)

msg = ('JaxArray cannot be updated in JIT mode. You should '
       'mark it as the brainpy.math.Variable instead.')


class JaxArray(object):
  """Multiple-dimensional array for JAX backend.
  """
  __slots__ = ("_value", "_jit_mode")

  def __init__(self, value):
    if isinstance(value, (list, tuple)):
      value = jnp.asarray(value)
    if isinstance(value, JaxArray):
      value = value._value
    self._value = value
    self._jit_mode = False

  @property
  def value(self):
    return self._value

  @value.setter
  def value(self, value):
    self.update(value)

  def update(self, value):
    """Update the value of this JaxArray.
    """
    if self._jit_mode:
      raise MathError(msg)
    if value.shape != self._value.shape:
      raise MathError(f"The shape of the original data is {self._value.shape}, "
                      f"while we got {value.shape}.")
    if value.dtype != self._value.dtype:
      raise MathError(f"The dtype of the original data is {self._value.dtype}, "
                      f"while we got {value.dtype}.")
    self._value = value.value if isinstance(value, JaxArray) else value

  @property
  def dtype(self):
    return self._value.dtype

  @property
  def shape(self):
    return self._value.shape

  @property
  def ndim(self):
    return self._value.ndim

  @property
  def imag(self):
    return self._value.image

  @property
  def real(self):
    return JaxArray(self._value.real)

  @property
  def size(self):
    return self.value.size

  @property
  def T(self):
    return JaxArray(self.value.T)

  # ----------------------- #
  # Python inherent methods #
  # ----------------------- #

  def __repr__(self) -> str:
    print_code = repr(self.value)
    name = self.__class__.__name__
    if 'DeviceArray' in print_code:
      print_code = print_code.replace('DeviceArray', name)
      lines = print_code.split("\n")
      if len(name) > len('DeviceArray'):
        num_len = len(name) - len('DeviceArray')
        for i in range(1, len(lines)):
          lines[i] = " " * num_len + lines[i]
      else:
        num_len = len('DeviceArray') - len(name)
        for i in range(1, len(lines)):
          lines[i] = lines[i][num_len:]
      print_code = "\n".join(lines)
    else:
      lines = print_code.split("\n")
      prefix = name + "("
      lines[0] = prefix + lines[0]
      prefix = " " * len(prefix)
      for i in range(1, len(lines)):
        lines[i] = prefix + lines[i]
      lines[-1] = lines[-1] + ")"
      print_code = "\n".join(lines)
    return print_code

  def __format__(self, format_spec: str) -> str:
    return format(self.value)

  def __iter__(self):
    """Solve the issue of DeviceArray.__iter__.

    Details please see JAX issues:

    - https://github.com/google/jax/issues/7713
    - https://github.com/google/jax/pull/3821
    """
    for i in range(self._value.shape[0]):
      yield self._value[i]

  def __getitem__(self, index):
    if isinstance(index, slice) and (index == _all_slice):
      return self.value
    elif isinstance(index, tuple):
      index = tuple(x.value if isinstance(x, JaxArray) else x for x in index)
    elif isinstance(index, JaxArray):
      index = index.value
    return self.value[index]

  def __setitem__(self, index, value):
    if self._jit_mode:
      raise MathError(msg)

    # value is JaxArray
    if isinstance(value, JaxArray):
      value = value.value

    # tuple index
    if isinstance(index, tuple):
      index = tuple(x.value if isinstance(x, JaxArray) else x for x in index)

    # JaxArray index
    elif isinstance(index, JaxArray):
      index = index.value

    # update
    self._value = self._value.at[index].set(value)

  # ---------- #
  # operations #
  # ---------- #

  def __bool__(self) -> bool:
    return bool(self._value)

  def __len__(self) -> int:
    return len(self._value)

  def __neg__(self):
    return JaxArray(self._value.__neg__())

  def __pos__(self):
    return JaxArray(self._value.__pos__())

  def __abs__(self):
    return JaxArray(self._value.__abs__())

  def __invert__(self):
    return JaxArray(self._value.__invert__())

  def __eq__(self, oc):
    return JaxArray(self._value.__eq__(oc._value if isinstance(oc, JaxArray) else oc))

  def __ne__(self, oc):
    return JaxArray(self._value.__ne__(oc._value if isinstance(oc, JaxArray) else oc))

  def __lt__(self, oc):
    return JaxArray(self._value.__lt__(oc._value if isinstance(oc, JaxArray) else oc))

  def __le__(self, oc):
    return JaxArray(self._value.__le__(oc._value if isinstance(oc, JaxArray) else oc))

  def __gt__(self, oc):
    return JaxArray(self._value.__gt__(oc._value if isinstance(oc, JaxArray) else oc))

  def __ge__(self, oc):
    return JaxArray(self._value.__ge__(oc._value if isinstance(oc, JaxArray) else oc))

  def __add__(self, oc):
    return JaxArray(self._value.__add__(oc._value if isinstance(oc, JaxArray) else oc))

  def __radd__(self, oc):
    return JaxArray(self._value.__radd__(oc._value if isinstance(oc, JaxArray) else oc))

  def __iadd__(self, oc):
    # a += b
    if self._jit_mode:
      raise MathError(msg)
    self._value += (oc._value if isinstance(oc, JaxArray) else oc)
    return self

  def __sub__(self, oc):
    return JaxArray(self._value.__sub__(oc._value if isinstance(oc, JaxArray) else oc))

  def __rsub__(self, oc):
    return JaxArray(self._value.__rsub__(oc._value if isinstance(oc, JaxArray) else oc))

  def __isub__(self, oc):
    # a -= b
    if self._jit_mode:
      raise MathError(msg)
    self._value = self._value.__sub__(oc._value if isinstance(oc, JaxArray) else oc)
    return self

  def __mul__(self, oc):
    return JaxArray(self._value.__mul__(oc._value if isinstance(oc, JaxArray) else oc))

  def __rmul__(self, oc):
    return JaxArray(self._value.__rmul__(oc._value if isinstance(oc, JaxArray) else oc))

  def __imul__(self, oc):
    # a *= b
    if self._jit_mode:
      raise MathError(msg)
    self._value = self._value.__mul__(oc._value if isinstance(oc, JaxArray) else oc)
    return self

  def __div__(self, oc):
    return JaxArray(self._value.__div__(oc._value if isinstance(oc, JaxArray) else oc))

  def __rdiv__(self, oc):
    return JaxArray(self._value.__rdiv__(oc._value if isinstance(oc, JaxArray) else oc))

  def __truediv__(self, oc):
    return JaxArray(self._value.__truediv__(oc._value if isinstance(oc, JaxArray) else oc))

  def __rtruediv__(self, oc):
    return JaxArray(self._value.__rtruediv__(oc._value if isinstance(oc, JaxArray) else oc))

  def __itruediv__(self, oc):
    # a /= b
    if self._jit_mode:
      raise MathError(msg)
    self._value = self._value.__truediv__(oc._value if isinstance(oc, JaxArray) else oc)
    return self

  def __floordiv__(self, oc):
    return JaxArray(self._value.__floordiv__(oc._value if isinstance(oc, JaxArray) else oc))

  def __rfloordiv__(self, oc):
    return JaxArray(self._value.__rfloordiv__(oc._value if isinstance(oc, JaxArray) else oc))

  def __ifloordiv__(self, oc):
    # a //= b
    if self._jit_mode:
      raise MathError(msg)
    self._value = self._value.__floordiv__(oc._value if isinstance(oc, JaxArray) else oc)
    return self

  def __divmod__(self, oc):
    return JaxArray(self._value.__divmod__(oc._value if isinstance(oc, JaxArray) else oc))

  def __rdivmod__(self, oc):
    return JaxArray(self._value.__rdivmod__(oc._value if isinstance(oc, JaxArray) else oc))

  def __mod__(self, oc):
    return JaxArray(self._value.__mod__(oc._value if isinstance(oc, JaxArray) else oc))

  def __rmod__(self, oc):
    return JaxArray(self._value.__rmod__(oc._value if isinstance(oc, JaxArray) else oc))

  def __imod__(self, oc):
    # a %= b
    if self._jit_mode:
      raise MathError(msg)
    self._value = self._value.__mod__(oc._value if isinstance(oc, JaxArray) else oc)
    return self

  def __pow__(self, oc):
    return JaxArray(self._value.__pow__(oc._value if isinstance(oc, JaxArray) else oc))

  def __rpow__(self, oc):
    return JaxArray(self._value.__rpow__(oc._value if isinstance(oc, JaxArray) else oc))

  def __ipow__(self, oc):
    # a **= b
    if self._jit_mode:
      raise MathError(msg)
    self._value = self._value ** (oc._value if isinstance(oc, JaxArray) else oc)
    return self

  def __matmul__(self, oc):
    return JaxArray(self._value.__matmul__(oc._value if isinstance(oc, JaxArray) else oc))

  def __rmatmul__(self, oc):
    return JaxArray(self._value.__rmatmul__(oc._value if isinstance(oc, JaxArray) else oc))

  def __imatmul__(self, oc):
    # a @= b
    if self._jit_mode:
      raise MathError(msg)
    self._value = self._value.__matmul__(oc._value if isinstance(oc, JaxArray) else oc)
    return self

  def __and__(self, oc):
    return JaxArray(self._value.__and__(oc._value if isinstance(oc, JaxArray) else oc))

  def __rand__(self, oc):
    return JaxArray(self._value.__rand__(oc._value if isinstance(oc, JaxArray) else oc))

  def __iand__(self, oc):
    # a &= b
    if self._jit_mode:
      raise MathError(msg)
    self._value = self._value.__and__(oc._value if isinstance(oc, JaxArray) else oc)
    return self

  def __or__(self, oc):
    return JaxArray(self._value.__or__(oc._value if isinstance(oc, JaxArray) else oc))

  def __ror__(self, oc):
    return JaxArray(self._value.__ror__(oc._value if isinstance(oc, JaxArray) else oc))

  def __ior__(self, oc):
    # a |= b
    if self._jit_mode:
      raise MathError(msg)
    self._value = self._value.__or__(oc._value if isinstance(oc, JaxArray) else oc)
    return self

  def __xor__(self, oc):
    return JaxArray(self._value.__xor__(oc._value if isinstance(oc, JaxArray) else oc))

  def __rxor__(self, oc):
    return JaxArray(self._value.__rxor__(oc._value if isinstance(oc, JaxArray) else oc))

  def __ixor__(self, oc):
    # a ^= b
    if self._jit_mode:
      raise MathError(msg)
    self._value = self._value.__xor__(oc._value if isinstance(oc, JaxArray) else oc)
    return self

  def __lshift__(self, oc):
    return JaxArray(self._value.__lshift__(oc._value if isinstance(oc, JaxArray) else oc))

  def __rlshift__(self, oc):
    return JaxArray(self._value.__rlshift__(oc._value if isinstance(oc, JaxArray) else oc))

  def __ilshift__(self, oc):
    # a <<= b
    if self._jit_mode:
      raise MathError(msg)
    self._value = self._value.__lshift__(oc._value if isinstance(oc, JaxArray) else oc)
    return self

  def __rshift__(self, oc):
    return JaxArray(self._value.__rshift__(oc._value if isinstance(oc, JaxArray) else oc))

  def __rrshift__(self, oc):
    return JaxArray(self._value.__rrshift__(oc._value if isinstance(oc, JaxArray) else oc))

  def __irshift__(self, oc):
    # a >>= b
    if self._jit_mode:
      raise MathError(msg)
    self._value = self._value.__rshift__(oc._value if isinstance(oc, JaxArray) else oc)
    return self

  def __round__(self, ndigits=None):
    return JaxArray(self._value.__round__(ndigits))

  # ----------------------- #
  #       JAX methods       #
  # ----------------------- #

  def block_host_until_ready(self, *args):
    self._value.block_host_until_ready(*args)

  def block_until_ready(self, *args):
    self._value.block_until_ready(*args)

  # def broadcast(self, operand, sizes):
  #   """Broadcasts an array, adding new major dimensions.
  #
  #   Wraps XLA's `Broadcast
  #   <https://www.tensorflow.org/xla/operation_semantics#broadcast>`_
  #   operator.
  #
  #   Parameters
  #   ----------
  #   operand: an array
  #   sizes:
  #     A sequence of integers, giving the sizes of new major dimensions
  #     to add.
  #
  #   Returns
  #   -------
  #   ary : array
  #     An array containing the result.
  #   """
  #   raise NotImplementedError
  #
  # def client(self, *args):
  #   raise NotImplementedError
  #
  # def clone(self, *args):
  #   raise NotImplementedError
  #
  # def copy_to_device(self, *args):
  #   raise NotImplementedError
  #
  # def copy_to_host_async(self, *args):
  #   raise NotImplementedError
  #
  # def device(self, *args):
  #   raise NotImplementedError
  #
  # def device_buffer(self, *args):
  #   raise NotImplementedError

  # ----------------------- #
  #      NumPy methods      #
  # ----------------------- #

  def all(self, axis=None, keepdims=False):
    """Returns True if all elements evaluate to True."""
    r = self.value.all(axis=axis, keepdims=keepdims)
    return r if (axis is None or keepdims) else JaxArray(r)

  def any(self, axis=None, keepdims=False):
    """Returns True if any of the elements of a evaluate to True."""
    r = self.value.any(axis=axis, keepdims=keepdims)
    return r if (axis is None or keepdims) else JaxArray(r)

  def argmax(self, axis=None):
    """Return indices of the maximum values along the given axis."""
    return JaxArray(self.value.argmax(axis=axis))

  def argmin(self, axis=None):
    """Return indices of the minimum values along the given axis."""
    return JaxArray(self.value.argmin(axis=axis))

  def argpartition(self, kth, axis=-1, kind='introselect', order=None):
    """Returns the indices that would partition this array."""
    return JaxArray(self.value.argpartition(kth=kth, axis=axis, kind=kind, order=order))

  def argsort(self, axis=-1, kind=None, order=None):
    """Returns the indices that would sort this array."""
    return JaxArray(self.value.argsort(axis=axis, kind=kind, order=order))

  def astype(self, dtype):
    """Copy of the array, cast to a specified type.

    Parameters
    ----------
    dtype: str, dtype
      Typecode or data-type to which the array is cast.
    """
    return JaxArray(self.value.astype(dtype=dtype))

  def byteswap(self, inplace=False):
    """Swap the bytes of the array elements

    Toggle between low-endian and big-endian data representation by
    returning a byteswapped array, optionally swapped in-place.
    Arrays of byte-strings are not swapped. The real and imaginary
    parts of a complex number are swapped individually."""
    return JaxArray(self.value.byteswap(inplace=inplace))

  def choose(self, choices, mode='raise'):
    """Use an index array to construct a new array from a set of choices."""
    choices = choices.value if isinstance(choices, JaxArray) else choices
    return JaxArray(self.value.choose(choices=choices, mode=mode))

  def clip(self, min=None, max=None):
    """Return an array whose values are limited to [min, max]. One of max or min must be given."""
    return JaxArray(self.value.clip(min=min, max=max))

  def compress(self, condition, axis=None):
    """Return selected slices of this array along given axis."""
    condition = condition.value if isinstance(condition, JaxArray) else condition
    return JaxArray(self.value.compress(condition=condition, axis=axis))

  def conj(self):
    """Complex-conjugate all elements."""
    return JaxArray(self.value.conj())

  def conjugate(self):
    """Return the complex conjugate, element-wise."""
    return JaxArray(self.value.conjugate())

  def copy(self):
    """Return a copy of the array."""
    return JaxArray(self.value.copy())

  def cumprod(self, axis=None, dtype=None):
    """Return the cumulative product of the elements along the given axis."""
    return JaxArray(self.value.cumprod(axis=axis, dtype=dtype))

  def cumsum(self, axis=None, dtype=None):
    """Return the cumulative sum of the elements along the given axis."""
    return JaxArray(self.value.cumsum(axis=axis, dtype=dtype))

  def diagonal(self, offset=0, axis1=0, axis2=1):
    """Return specified diagonals."""
    return JaxArray(self.value.diagonal(offset=offset, axis1=axis1, axis2=axis2))

  def dot(self, b):
    """Dot product of two arrays."""
    return JaxArray(self.value.dot(b))

  # def dump(self, file):
  #   """Dump a pickle of the array to the specified file.
  #   The array can be read back with pickle.load or numpy.load."""
  #   pass

  # def dumps(self):
  #   """Returns the pickle of the array as a string. pickle.loads
  #   or numpy.loads will convert the string back to an array."""
  #   pass

  def fill(self, value):
    """Fill the array with a scalar value."""
    self._value = jnp.ones_like(self.value) * value

  def flatten(self, order='C'):
    return JaxArray(self.value.flatten(order=order))

  # def getfield(self, dtype, offset=0):
  #   return type(self)(self.value.getfield(dtype=dtype, offset=offset))

  def item(self, *args):
    """Copy an element of an array to a standard Python scalar and return it."""
    return self.value.item(*args)

  # def itemset(self, *args):
  #   """Insert scalar into an array (scalar is cast to array’s dtype, if possible)"""
  #   self.value.itemset(*args)

  def max(self, axis=None, keepdims=False, *args, **kwargs):
    """Return the maximum along a given axis."""
    res = self.value.max(axis=axis, keepdims=keepdims, *args, **kwargs)
    return res if (axis is None or keepdims) else JaxArray(res)

  def mean(self, axis=None, dtype=None, keepdims=False, *args, **kwargs):
    """Returns the average of the array elements along given axis."""
    res = self.value.mean(axis=axis, dtype=dtype, keepdims=keepdims, *args, **kwargs)
    return res if (axis is None or keepdims) else JaxArray(res)

  def min(self, axis=None, keepdims=False, *args, **kwargs):
    """Return the minimum along a given axis."""
    res = self.value.min(axis=axis, keepdims=keepdims, *args, **kwargs)
    return res if (axis is None or keepdims) else JaxArray(res)

  def nonzero(self):
    """Return the indices of the elements that are non-zero."""
    return tuple(JaxArray(a) for a in self.value.nonzero())

  def prod(self, axis=None, dtype=None, keepdims=False, initial=1, where=True):
    """Return the product of the array elements over the given axis."""
    res = self.value.prod(axis=axis, dtype=dtype, keepdims=keepdims, initial=initial, where=where)
    return res if (axis is None or keepdims) else JaxArray(res)

  def ptp(self, axis=None, keepdims=False):
    """Peak to peak (maximum - minimum) value along a given axis."""
    r = self.value.ptp(axis=axis, keepdims=keepdims)
    return r if (axis is None or keepdims) else JaxArray(r)

  def ravel(self, order=None):
    """Return a flattened array."""
    return JaxArray(self.value.ravel(order=order))

  def repeat(self, repeats, axis=None):
    """Repeat elements of an array."""
    return JaxArray(self.value.repeat(repeats=repeats, axis=axis))

  def reshape(self, shape, order='C'):
    """Returns an array containing the same data with a new shape."""
    return JaxArray(self.value.reshape(*shape, order=order))

  def round(self, decimals=0):
    """Return ``a`` with each element rounded to the given number of decimals."""
    return JaxArray(self.value.round(decimals=decimals))

  def searchsorted(self, v, side='left', sorter=None):
    """Find indices where elements should be inserted to maintain order.

    Find the indices into a sorted array `a` such that, if the
    corresponding elements in `v` were inserted before the indices, the
    order of `a` would be preserved.

    Assuming that `a` is sorted:

    ======  ============================
    `side`  returned index `i` satisfies
    ======  ============================
    left    ``a[i-1] < v <= a[i]``
    right   ``a[i-1] <= v < a[i]``
    ======  ============================

    Parameters
    ----------
    v : array_like
        Values to insert into `a`.
    side : {'left', 'right'}, optional
        If 'left', the index of the first suitable location found is given.
        If 'right', return the last such index.  If there is no suitable
        index, return either 0 or N (where N is the length of `a`).
    sorter : 1-D array_like, optional
        Optional array of integer indices that sort array a into ascending
        order. They are typically the result of argsort.

    Returns
    -------
    indices : array of ints
        Array of insertion points with the same shape as `v`.
    """
    v = v.value if isinstance(v, JaxArray) else v
    return JaxArray(self.value.searchsorted(v=v, side=side, sorter=sorter))

  def sort(self, axis=-1, kind=None, order=None):
    """Sort an array in-place.

    Parameters
    ----------
    axis : int, optional
        Axis along which to sort. Default is -1, which means sort along the
        last axis.
    kind : {'quicksort', 'mergesort', 'heapsort', 'stable'}, optional
        Sorting algorithm. The default is 'quicksort'. Note that both 'stable'
        and 'mergesort' use timsort under the covers and, in general, the
        actual implementation will vary with datatype. The 'mergesort' option
        is retained for backwards compatibility.
    order : str or list of str, optional
        When `a` is an array with fields defined, this argument specifies
        which fields to compare first, second, etc.  A single field can
        be specified as a string, and not all fields need be specified,
        but unspecified fields will still be used, in the order in which
        they come up in the dtype, to break ties.
    """
    self._value = self.value.sort(axis=axis, kind=kind, order=order)

  def squeeze(self, axis=None):
    """Remove axes of length one from ``a``."""
    return JaxArray(self.value.squeeze(axis=axis))

  def std(self, axis=None, dtype=None, ddof=0, keepdims=False):
    """Compute the standard deviation along the specified axis.

    Returns the standard deviation, a measure of the spread of a distribution,
    of the array elements. The standard deviation is computed for the
    flattened array by default, otherwise over the specified axis.

    Parameters
    ----------
    axis : None or int or tuple of ints, optional
        Axis or axes along which the standard deviation is computed. The
        default is to compute the standard deviation of the flattened array.
        If this is a tuple of ints, a standard deviation is performed over
        multiple axes, instead of a single axis or all the axes as before.
    dtype : dtype, optional
        Type to use in computing the standard deviation. For arrays of
        integer type the default is float64, for arrays of float types it is
        the same as the array type.
    out : ndarray, optional
        Alternative output array in which to place the result. It must have
        the same shape as the expected output but the type (of the calculated
        values) will be cast if necessary.
    ddof : int, optional
        Means Delta Degrees of Freedom.  The divisor used in calculations
        is ``N - ddof``, where ``N`` represents the number of elements.
        By default `ddof` is zero.
    keepdims : bool, optional
        If this is set to True, the axes which are reduced are left
        in the result as dimensions with size one. With this option,
        the result will broadcast correctly against the input array.

        If the default value is passed, then `keepdims` will not be
        passed through to the `std` method of sub-classes of
        `ndarray`, however any non-default value will be.  If the
        sub-class' method does not implement `keepdims` any
        exceptions will be raised.
    where : array_like of bool, optional
        Elements to include in the standard deviation.

    Returns
    -------
    standard_deviation : ndarray, see dtype parameter above.
        If `out` is None, return a new array containing the standard deviation,
        otherwise return a reference to the output array.
    """
    r = self.value.std(axis=axis, dtype=dtype, ddof=ddof, keepdims=keepdims)
    return r if (axis is None or keepdims) else JaxArray(r)

  def sum(self, axis=None, dtype=None, keepdims=False, initial=0, where=True):
    """Return the sum of the array elements over the given axis."""
    res = self.value.sum(axis=axis, dtype=dtype, keepdims=keepdims, initial=initial, where=where)
    return res if (axis is None or keepdims) else JaxArray(res)

  def swapaxes(self, axis1, axis2):
    """Return a view of the array with `axis1` and `axis2` interchanged."""
    return JaxArray(self.value.swapaxes(axis1, axis2))

  def split(self, indices_or_sections, axis=0):
    """Split an array into multiple sub-arrays as views into ``ary``.

    Parameters
    ----------
    indices_or_sections : int, 1-D array
      If `indices_or_sections` is an integer, N, the array will be divided
      into N equal arrays along `axis`.  If such a split is not possible,
      an error is raised.

      If `indices_or_sections` is a 1-D array of sorted integers, the entries
      indicate where along `axis` the array is split.  For example,
      ``[2, 3]`` would, for ``axis=0``, result in

        - ary[:2]
        - ary[2:3]
        - ary[3:]

      If an index exceeds the dimension of the array along `axis`,
      an empty sub-array is returned correspondingly.
    axis : int, optional
      The axis along which to split, default is 0.

    Returns
    -------
    sub-arrays : list of ndarrays
      A list of sub-arrays as views into `ary`.
    """
    return [JaxArray(a) for a in self.value.split(indices_or_sections, axis=axis)]

  def take(self, indices, axis=None, mode=None):
    """Return an array formed from the elements of a at the given indices."""
    indices = indices.value if isinstance(indices, JaxArray) else indices
    return JaxArray(self.value.take(indices=indices, axis=axis, mode=mode))

  def tobytes(self, order='C'):
    """Construct Python bytes containing the raw data bytes in the array.

    Constructs Python bytes showing a copy of the raw contents of data memory.
    The bytes object is produced in C-order by default. This behavior is
    controlled by the ``order`` parameter."""
    return JaxArray(self.value.tobytes(order=order))

  def tolist(self):
    """Return the array as an ``a.ndim``-levels deep nested list of Python scalars.

    Return a copy of the array data as a (nested) Python list.
    Data items are converted to the nearest compatible builtin Python type, via
    the `~numpy.ndarray.item` function.

    If ``a.ndim`` is 0, then since the depth of the nested list is 0, it will
    not be a list at all, but a simple Python scalar.
    """
    return self.value.tolist()

  def trace(self, offset=0, axis1=0, axis2=1, dtype=None):
    """Return the sum along diagonals of the array."""
    return JaxArray(self.value.trace(offset=offset, axis1=axis1, axis2=axis2, dtype=dtype))

  def transpose(self, *axes):
    """Returns a view of the array with axes transposed.

    For a 1-D array this has no effect, as a transposed vector is simply the
    same vector. To convert a 1-D array into a 2D column vector, an additional
    dimension must be added. `np.atleast2d(a).T` achieves this, as does
    `a[:, np.newaxis]`.
    For a 2-D array, this is a standard matrix transpose.
    For an n-D array, if axes are given, their order indicates how the
    axes are permuted (see Examples). If axes are not provided and
    ``a.shape = (i[0], i[1], ... i[n-2], i[n-1])``, then
    ``a.transpose().shape = (i[n-1], i[n-2], ... i[1], i[0])``.

    Parameters
    ----------
    axes : None, tuple of ints, or `n` ints

     * None or no argument: reverses the order of the axes.

     * tuple of ints: `i` in the `j`-th place in the tuple means `a`'s
       `i`-th axis becomes `a.transpose()`'s `j`-th axis.

     * `n` ints: same as an n-tuple of the same ints (this form is
       intended simply as a "convenience" alternative to the tuple form)

    Returns
    -------
    out : ndarray
        View of `a`, with axes suitably permuted.
    """
    return JaxArray(self.value.transpose(*axes))

  def tile(self, reps):
    """Construct an array by repeating A the number of times given by reps.

    If `reps` has length ``d``, the result will have dimension of
    ``max(d, A.ndim)``.

    If ``A.ndim < d``, `A` is promoted to be d-dimensional by prepending new
    axes. So a shape (3,) array is promoted to (1, 3) for 2-D replication,
    or shape (1, 1, 3) for 3-D replication. If this is not the desired
    behavior, promote `A` to d-dimensions manually before calling this
    function.

    If ``A.ndim > d``, `reps` is promoted to `A`.ndim by pre-pending 1's to it.
    Thus for an `A` of shape (2, 3, 4, 5), a `reps` of (2, 2) is treated as
    (1, 1, 2, 2).

    Note : Although tile may be used for broadcasting, it is strongly
    recommended to use numpy's broadcasting operations and functions.

    Parameters
    ----------
    reps : array_like
        The number of repetitions of `A` along each axis.

    Returns
    -------
    c : ndarray
        The tiled output array.
    """
    reps = reps.value if isinstance(reps, JaxArray) else reps
    return JaxArray(self.value.tile(reps))

  def var(self, axis=None, dtype=None, ddof=0, keepdims=False):
    """Returns the variance of the array elements, along given axis."""
    r = self.value.var(axis=axis, dtype=dtype, ddof=ddof, keepdims=keepdims)
    return r if (axis is None or keepdims) else JaxArray(r)

  def view(self, dtype=None, *args, **kwargs):
    """New view of array with the same data."""
    return JaxArray(self.value.view(dtype=dtype, *args, **kwargs))

  # ------------------
  # NumPy support
  # ------------------

  def numpy(self):
    """Convert to numpy.ndarray."""
    return np.asarray(self.value)

  def to_numpy(self):
    """Convert to numpy.ndarray."""
    return np.asarray(self.value)

  def to_jax(self):
    """Convert to jax.numpy.ndarray."""
    return self.value

  def __array__(self):
    """Support ``numpy.array()`` and ``numpy.asarray()`` functions."""
    return np.asarray(self.value)

  def __jax_array__(self):
    return self.value


ndarray = JaxArray


class Variable(JaxArray):
  """The pointer to specify the dynamical variable.
  """
  __slots__ = ('_value', )

  def __init__(self, value):
    super(Variable, self).__init__(value)

  @property
  def real(self):
    return self._value.real

  @property
  def T(self):
    return self.value.T

  def __neg__(self):
    return self._value.__neg__()

  def __pos__(self):
    return self._value.__pos__()

  def __abs__(self):
    return self._value.__abs__()

  def __invert__(self):
    return self._value.__invert__()

  def __eq__(self, oc):
    return self._value.__eq__(oc._value if isinstance(oc, JaxArray) else oc)

  def __ne__(self, oc):
    return self._value.__ne__(oc._value if isinstance(oc, JaxArray) else oc)

  def __lt__(self, oc):
    return self._value.__lt__(oc._value if isinstance(oc, JaxArray) else oc)

  def __le__(self, oc):
    return self._value.__le__(oc._value if isinstance(oc, JaxArray) else oc)

  def __gt__(self, oc):
    return self._value.__gt__(oc._value if isinstance(oc, JaxArray) else oc)

  def __ge__(self, oc):
    return self._value.__ge__(oc._value if isinstance(oc, JaxArray) else oc)

  def __add__(self, oc):
    return self._value.__add__(oc._value if isinstance(oc, JaxArray) else oc)

  def __radd__(self, oc):
    return self._value.__radd__(oc._value if isinstance(oc, JaxArray) else oc)

  def __iadd__(self, oc):
    # a += b
    self._value += (oc._value if isinstance(oc, JaxArray) else oc)
    return self

  def __sub__(self, oc):
    return self._value.__sub__(oc._value if isinstance(oc, JaxArray) else oc)

  def __rsub__(self, oc):
    return self._value.__rsub__(oc._value if isinstance(oc, JaxArray) else oc)

  def __isub__(self, oc):
    # a -= b
    self._value = self._value.__sub__(oc._value if isinstance(oc, JaxArray) else oc)
    return self

  def __mul__(self, oc):
    return self._value.__mul__(oc._value if isinstance(oc, JaxArray) else oc)

  def __rmul__(self, oc):
    return self._value.__rmul__(oc._value if isinstance(oc, JaxArray) else oc)

  def __imul__(self, oc):
    # a *= b
    self._value = self._value.__mul__(oc._value if isinstance(oc, JaxArray) else oc)
    return self

  def __div__(self, oc):
    return self._value.__div__(oc._value if isinstance(oc, JaxArray) else oc)

  def __rdiv__(self, oc):
    return self._value.__rdiv__(oc._value if isinstance(oc, JaxArray) else oc)

  def __truediv__(self, oc):
    return self._value.__truediv__(oc._value if isinstance(oc, JaxArray) else oc)

  def __rtruediv__(self, oc):
    return self._value.__rtruediv__(oc._value if isinstance(oc, JaxArray) else oc)

  def __itruediv__(self, oc):
    # a /= b
    self._value = self._value.__truediv__(oc._value if isinstance(oc, JaxArray) else oc)
    return self

  def __floordiv__(self, oc):
    return self._value.__floordiv__(oc._value if isinstance(oc, JaxArray) else oc)

  def __rfloordiv__(self, oc):
    return self._value.__rfloordiv__(oc._value if isinstance(oc, JaxArray) else oc)

  def __ifloordiv__(self, oc):
    # a //= b
    self._value = self._value.__floordiv__(oc._value if isinstance(oc, JaxArray) else oc)
    return self

  def __divmod__(self, oc):
    return self._value.__divmod__(oc._value if isinstance(oc, JaxArray) else oc)

  def __rdivmod__(self, oc):
    return self._value.__rdivmod__(oc._value if isinstance(oc, JaxArray) else oc)

  def __mod__(self, oc):
    return self._value.__mod__(oc._value if isinstance(oc, JaxArray) else oc)

  def __rmod__(self, oc):
    return self._value.__rmod__(oc._value if isinstance(oc, JaxArray) else oc)

  def __imod__(self, oc):
    # a %= b
    self._value = self._value.__mod__(oc._value if isinstance(oc, JaxArray) else oc)
    return self

  def __pow__(self, oc):
    return self._value.__pow__(oc._value if isinstance(oc, JaxArray) else oc)

  def __rpow__(self, oc):
    return self._value.__rpow__(oc._value if isinstance(oc, JaxArray) else oc)

  def __ipow__(self, oc):
    # a **= b
    self._value = self._value ** (oc._value if isinstance(oc, JaxArray) else oc)
    return self

  def __matmul__(self, oc):
    return self._value.__matmul__(oc._value if isinstance(oc, JaxArray) else oc)

  def __rmatmul__(self, oc):
    return self._value.__rmatmul__(oc._value if isinstance(oc, JaxArray) else oc)

  def __imatmul__(self, oc):
    # a @= b
    self._value = self._value.__matmul__(oc._value if isinstance(oc, JaxArray) else oc)
    return self

  def __and__(self, oc):
    return self._value.__and__(oc._value if isinstance(oc, JaxArray) else oc)

  def __rand__(self, oc):
    return self._value.__rand__(oc._value if isinstance(oc, JaxArray) else oc)
  def __iand__(self, oc):
    # a &= b
    self._value = self._value.__and__(oc._value if isinstance(oc, JaxArray) else oc)
    return self

  def __or__(self, oc):
    return self._value.__or__(oc._value if isinstance(oc, JaxArray) else oc)

  def __ror__(self, oc):
    return self._value.__ror__(oc._value if isinstance(oc, JaxArray) else oc)

  def __ior__(self, oc):
    # a |= b
    self._value = self._value.__or__(oc._value if isinstance(oc, JaxArray) else oc)
    return self

  def __xor__(self, oc):
    return self._value.__xor__(oc._value if isinstance(oc, JaxArray) else oc)

  def __rxor__(self, oc):
    return self._value.__rxor__(oc._value if isinstance(oc, JaxArray) else oc)

  def __ixor__(self, oc):
    # a ^= b
    self._value = self._value.__xor__(oc._value if isinstance(oc, JaxArray) else oc)
    return self

  def __lshift__(self, oc):
    return self._value.__lshift__(oc._value if isinstance(oc, JaxArray) else oc)

  def __rlshift__(self, oc):
    return self._value.__rlshift__(oc._value if isinstance(oc, JaxArray) else oc)

  def __ilshift__(self, oc):
    # a <<= b
    self._value = self._value.__lshift__(oc._value if isinstance(oc, JaxArray) else oc)
    return self

  def __rshift__(self, oc):
    return self._value.__rshift__(oc._value if isinstance(oc, JaxArray) else oc)

  def __rrshift__(self, oc):
    return self._value.__rrshift__(oc._value if isinstance(oc, JaxArray) else oc)

  def __irshift__(self, oc):
    # a >>= b
    self._value = self._value.__rshift__(oc._value if isinstance(oc, JaxArray) else oc)
    return self

  def __round__(self, ndigits=None):
    return self._value.__round__(ndigits)

  # ----------------------- #
  #       JAX methods       #
  # ----------------------- #

  def block_host_until_ready(self, *args):
    self._value.block_host_until_ready(*args)

  def block_until_ready(self, *args):
    self._value.block_until_ready(*args)

  # ----------------------- #
  #      NumPy methods      #
  # ----------------------- #

  def all(self, axis=None, keepdims=False):
    """Returns True if all elements evaluate to True."""
    r = self.value.all(axis=axis, keepdims=keepdims)
    return r

  def any(self, axis=None, keepdims=False):
    """Returns True if any of the elements of a evaluate to True."""
    r = self.value.any(axis=axis, keepdims=keepdims)
    return r

  def argmax(self, axis=None):
    """Return indices of the maximum values along the given axis."""
    return self.value.argmax(axis=axis)

  def argmin(self, axis=None):
    """Return indices of the minimum values along the given axis."""
    return self.value.argmin(axis=axis)

  def argpartition(self, kth, axis=-1, kind='introselect', order=None):
    """Returns the indices that would partition this array."""
    return self.value.argpartition(kth=kth, axis=axis, kind=kind, order=order)

  def argsort(self, axis=-1, kind=None, order=None):
    """Returns the indices that would sort this array."""
    return self.value.argsort(axis=axis, kind=kind, order=order)

  def astype(self, dtype):
    """Copy of the array, cast to a specified type.

    Parameters
    ----------
    dtype: str, dtype
      Typecode or data-type to which the array is cast.
    """
    return JaxArray(self.value.astype(dtype=dtype))

  def byteswap(self, inplace=False):
    """Swap the bytes of the array elements

    Toggle between low-endian and big-endian data representation by
    returning a byteswapped array, optionally swapped in-place.
    Arrays of byte-strings are not swapped. The real and imaginary
    parts of a complex number are swapped individually."""
    return self.value.byteswap(inplace=inplace)

  def choose(self, choices, mode='raise'):
    """Use an index array to construct a new array from a set of choices."""
    choices = choices.value if isinstance(choices, JaxArray) else choices
    return self.value.choose(choices=choices, mode=mode)

  def clip(self, min=None, max=None):
    """Return an array whose values are limited to [min, max]. One of max or min must be given."""
    return self.value.clip(min=min, max=max)

  def compress(self, condition, axis=None):
    """Return selected slices of this array along given axis."""
    condition = condition.value if isinstance(condition, JaxArray) else condition
    return self.value.compress(condition=condition, axis=axis)

  def conj(self):
    """Complex-conjugate all elements."""
    return self.value.conj()

  def conjugate(self):
    """Return the complex conjugate, element-wise."""
    return self.value.conjugate()

  def copy(self):
    """Return a copy of the array."""
    return self.value.copy()

  def cumprod(self, axis=None, dtype=None):
    """Return the cumulative product of the elements along the given axis."""
    return self.value.cumprod(axis=axis, dtype=dtype)

  def cumsum(self, axis=None, dtype=None):
    """Return the cumulative sum of the elements along the given axis."""
    return self.value.cumsum(axis=axis, dtype=dtype)

  def diagonal(self, offset=0, axis1=0, axis2=1):
    """Return specified diagonals."""
    return self.value.diagonal(offset=offset, axis1=axis1, axis2=axis2)

  def dot(self, b):
    """Dot product of two arrays."""
    return self.value.dot(b)

  def fill(self, value):
    """Fill the array with a scalar value."""
    self._value = jnp.ones_like(self.value) * value

  def flatten(self, order='C'):
    return self.value.flatten(order=order)

  def item(self, *args):
    """Copy an element of an array to a standard Python scalar and return it."""
    return self.value.item(*args)

  def max(self, axis=None, keepdims=False, *args, **kwargs):
    """Return the maximum along a given axis."""
    res = self.value.max(axis=axis, keepdims=keepdims, *args, **kwargs)
    return res

  def mean(self, axis=None, dtype=None, keepdims=False, *args, **kwargs):
    """Returns the average of the array elements along given axis."""
    res = self.value.mean(axis=axis, dtype=dtype, keepdims=keepdims, *args, **kwargs)
    return res

  def min(self, axis=None, keepdims=False, *args, **kwargs):
    """Return the minimum along a given axis."""
    res = self.value.min(axis=axis, keepdims=keepdims, *args, **kwargs)
    return res

  def nonzero(self):
    """Return the indices of the elements that are non-zero."""
    return self.value.nonzero()

  def prod(self, axis=None, dtype=None, keepdims=False, initial=1, where=True):
    """Return the product of the array elements over the given axis."""
    res = self.value.prod(axis=axis, dtype=dtype, keepdims=keepdims, initial=initial, where=where)
    return res

  def ptp(self, axis=None, keepdims=False):
    """Peak to peak (maximum - minimum) value along a given axis."""
    r = self.value.ptp(axis=axis, keepdims=keepdims)
    return r

  def ravel(self, order=None):
    """Return a flattened array."""
    return self.value.ravel(order=order)

  def repeat(self, repeats, axis=None):
    """Repeat elements of an array."""
    return self.value.repeat(repeats=repeats, axis=axis)

  def reshape(self, shape, order='C'):
    """Returns an array containing the same data with a new shape."""
    return self.value.reshape(*shape, order=order)

  def round(self, decimals=0):
    """Return ``a`` with each element rounded to the given number of decimals."""
    return self.value.round(decimals=decimals)

  def searchsorted(self, v, side='left', sorter=None):
    """Find indices where elements should be inserted to maintain order."""
    v = v.value if isinstance(v, JaxArray) else v
    return self.value.searchsorted(v=v, side=side, sorter=sorter)

  def sort(self, axis=-1, kind=None, order=None):
    """Sort an array in-place."""
    self._value = self.value.sort(axis=axis, kind=kind, order=order)

  def squeeze(self, axis=None):
    """Remove axes of length one from ``a``."""
    return self.value.squeeze(axis=axis)

  def std(self, axis=None, dtype=None, ddof=0, keepdims=False):
    """Compute the standard deviation along the specified axis.

    Returns the standard deviation, a measure of the spread of a distribution,
    of the array elements. The standard deviation is computed for the
    flattened array by default, otherwise over the specified axis.
    """
    r = self.value.std(axis=axis, dtype=dtype, ddof=ddof, keepdims=keepdims)
    return r

  def sum(self, axis=None, dtype=None, keepdims=False, initial=0, where=True):
    """Return the sum of the array elements over the given axis."""
    res = self.value.sum(axis=axis, dtype=dtype, keepdims=keepdims, initial=initial, where=where)
    return res

  def swapaxes(self, axis1, axis2):
    """Return a view of the array with `axis1` and `axis2` interchanged."""
    return self.value.swapaxes(axis1, axis2)

  def split(self, indices_or_sections, axis=0):
    """Split an array into multiple sub-arrays as views into ``ary``.
    """
    return self.value.split(indices_or_sections, axis=axis)

  def take(self, indices, axis=None, mode=None):
    """Return an array formed from the elements of a at the given indices."""
    indices = indices.value if isinstance(indices, JaxArray) else indices
    return self.value.take(indices=indices, axis=axis, mode=mode)

  def tobytes(self, order='C'):
    """Construct Python bytes containing the raw data bytes in the array.

    Constructs Python bytes showing a copy of the raw contents of data memory.
    The bytes object is produced in C-order by default. This behavior is
    controlled by the ``order`` parameter."""
    return self.value.tobytes(order=order)

  def tolist(self):
    """Return the array as an ``a.ndim``-levels deep nested list of Python scalars.

    Return a copy of the array data as a (nested) Python list.
    Data items are converted to the nearest compatible builtin Python type, via
    the `~numpy.ndarray.item` function.

    If ``a.ndim`` is 0, then since the depth of the nested list is 0, it will
    not be a list at all, but a simple Python scalar.
    """
    return self.value.tolist()

  def trace(self, offset=0, axis1=0, axis2=1, dtype=None):
    """Return the sum along diagonals of the array."""
    return self.value.trace(offset=offset, axis1=axis1, axis2=axis2, dtype=dtype)

  def transpose(self, *axes):
    """Returns a view of the array with axes transposed.
    """
    return self.value.transpose(*axes)

  def tile(self, reps):
    """Construct an array by repeating A the number of times given by reps.
    """
    reps = reps.value if isinstance(reps, JaxArray) else reps
    return self.value.tile(reps)

  def var(self, axis=None, dtype=None, ddof=0, keepdims=False):
    """Returns the variance of the array elements, along given axis."""
    r = self.value.var(axis=axis, dtype=dtype, ddof=ddof, keepdims=keepdims)
    return r

  def view(self, dtype=None, *args, **kwargs):
    """New view of array with the same data."""
    return self.value.view(dtype=dtype, *args, **kwargs)


class TrainVar(Variable):
  """The pointer to specify the trainable variable.
  """
  __slots__ = ('_value', )

  def __init__(self, value):
    super(TrainVar, self).__init__(value)


class Parameter(Variable):
  """The pointer to specify the parameter.
  """
  __slots__ = ('_value', )

  def __init__(self, value):
    super(Parameter, self).__init__(value)


register_pytree_node(JaxArray,
                     lambda t: ((t.value,), None),
                     lambda aux_data, flat_contents: JaxArray(*flat_contents))

register_pytree_node(Variable,
                     lambda t: ((t.value,), None),
                     lambda aux_data, flat_contents: Variable(*flat_contents))

register_pytree_node(TrainVar,
                     lambda t: ((t.value,), None),
                     lambda aux_data, flat_contents: TrainVar(*flat_contents))

register_pytree_node(Parameter,
                     lambda t: ((t.value,), None),
                     lambda aux_data, flat_contents: Parameter(*flat_contents))



# -*- coding: utf-8 -*-


import jax.lax
import jax.numpy as jnp
import numpy as np
from jax import grad, jit, vmap
from jax.flatten_util import ravel_pytree

import brainpy.math as bm
from brainpy import tools, errors
from . import f_without_jaxarray_return

try:
  import scipy.optimize as soptimize
except (ModuleNotFoundError, ImportError):
  soptimize = None


__all__ = [
  'ECONVERGED', 'ECONVERR',

  'jax_brentq',
  'get_brentq_candidates',
  'brentq_candidates',
  'brentq_roots',
  'brentq_roots2',
  'scipy_minimize_with_jax',
  'roots_of_1d_by_x',
  'roots_of_1d_by_xy',
]

ECONVERGED = 0
ECONVERR = -1


def _logical_or(a, b):
  a = a.value if isinstance(a, bm.JaxArray) else a
  b = b.value if isinstance(b, bm.JaxArray) else b
  return jnp.logical_or(a, b)


def _logical_and(a, b):
  a = a.value if isinstance(a, bm.JaxArray) else a
  b = b.value if isinstance(b, bm.JaxArray) else b
  return jnp.logical_and(a, b)


def _where(p, a, b):
  p = p.value if isinstance(p, bm.JaxArray) else p
  a = a.value if isinstance(a, bm.JaxArray) else a
  b = b.value if isinstance(b, bm.JaxArray) else b
  return jnp.where(p, a, b)


def jax_brentq(fun):
  f = f_without_jaxarray_return(fun)
  assert jax.config.read('jax_enable_x64'), ('Brentq optimization need x64 support. '
                                             'Please enable x64 with "brainpy.math.enable_x64()"')
  rtol = 4 * jnp.finfo(jnp.float64).eps
  # if jax.config.read('jax_enable_x64'):
  #   rtol = 4 * jnp.finfo(jnp.float64).eps
  # else:
  #   rtol = 1.5 * jnp.finfo(jnp.float32).eps

  def x(a, b, args=(), xtol=2e-14, maxiter=200):
    # Convert to float
    xpre = a * 1.0
    xcur = b * 1.0

    # Conditional checks for intervals in methods involving bisection
    fpre = f(xpre, *args)
    fcur = f(xcur, *args)

    # Root found at either end of [a,b]
    root = _where(fpre == 0, xpre, 0.)
    status = _where(fpre == 0, ECONVERGED, ECONVERR)
    root = _where(fcur == 0, xcur, root)
    status = _where(fcur == 0, ECONVERGED, status)

    # Check for sign error and early termination
    # Perform Brent's method
    def _f1(x):
      x['xblk'] = x['xpre']
      x['fblk'] = x['fpre']
      x['spre'] = x['xcur'] - x['xpre']
      x['scur'] = x['xcur'] - x['xpre']
      return x

    def _f2(x):
      x['xpre'] = x['xcur']
      x['xcur'] = x['xblk']
      x['xblk'] = x['xpre']
      x['fpre'] = x['fcur']
      x['fcur'] = x['fblk']
      x['fblk'] = x['fpre']
      return x

    def _f5(x):
      x['stry'] = -x['fcur'] * (x['xcur'] - x['xpre']) / (x['fcur'] - x['fpre'])
      return x

    def _f6(x):
      x['dpre'] = (x['fpre'] - x['fcur']) / (x['xpre'] - x['xcur'])
      dblk = (x['fblk'] - x['fcur']) / (x['xblk'] - x['xcur'])
      _tmp = dblk * x['dpre'] * (x['fblk'] - x['fpre'])
      x['stry'] = -x['fcur'] * (x['fblk'] * dblk - x['fpre'] * x['dpre']) / _tmp
      return x

    def _f3(x):
      x = jax.lax.cond(x['xpre'] == x['xblk'], _f5, _f6, x)
      k = jnp.min(jnp.array([abs(x['spre']), 3 * abs(x['sbis']) - x['delta']]))
      j = 2 * abs(x['stry']) < k
      x['spre'] = _where(j, x['scur'], x['sbis'])
      x['scur'] = _where(j, x['stry'], x['sbis'])
      return x

    def _f4(x):  # bisect
      x['spre'] = x['sbis']
      x['scur'] = x['sbis']
      return x

    def body_fun(x):
      x['itr'] += 1
      x = jax.lax.cond(x['fpre'] * x['fcur'] < 0, _f1, lambda a: a, x)
      x = jax.lax.cond(abs(x['fblk']) < abs(x['fcur']), _f2, lambda a: a, x)
      x['delta'] = (xtol + rtol * abs(x['xcur'])) / 2
      x['sbis'] = (x['xblk'] - x['xcur']) / 2
      # Root found
      j = _logical_or(x['fcur'] == 0, abs(x['sbis']) < x['delta'])
      x['status'] = _where(j, ECONVERGED, x['status'])
      x['root'] = _where(j, x['xcur'], x['root'])
      x = jax.lax.cond(_logical_and(abs(x['spre']) > x['delta'], abs(x['fcur']) < abs(x['fpre'])),
                       _f3, _f4, x)
      x['xpre'] = x['xcur']
      x['fpre'] = x['fcur']
      x['xcur'] += _where(abs(x['scur']) > x['delta'],
                          x['scur'], _where(x['sbis'] > 0, x['delta'], -x['delta']))
      x['fcur'] = f(x['xcur'], *args)
      x['funcalls'] += 1
      return x

    def cond_fun(R):
      return jnp.logical_and(R['status'] != ECONVERGED, R['itr'] <= maxiter)

    R = dict(root=root, status=status, xpre=xpre, xcur=xcur, fpre=fpre, fcur=fcur,
             itr=0, funcalls=2, xblk=xpre, fblk=fpre,
             sbis=(xpre - xcur) / 2,
             delta=(xtol + rtol * abs(xcur)) / 2,
             stry=-fcur * (xcur - xpre) / (fcur - fpre),
             scur=xcur - xpre, spre=xcur - xpre,
             dpre=(fpre - fcur) / (xpre - xcur))
    R = jax.lax.cond(status == ECONVERGED,
                     lambda x: x,
                     lambda x: jax.lax.while_loop(cond_fun, body_fun, x),
                     R)
    return dict(root=R['root'], funcalls=R['funcalls'], itr=R['itr'], status=R['status'])

  return x


def get_brentq_candidates(f, xs, ys):
  f = f_without_jaxarray_return(f)
  xs = xs.value if isinstance(xs, bm.JaxArray) else xs
  ys = ys.value if isinstance(ys, bm.JaxArray) else ys
  Y, X = jnp.meshgrid(ys, xs)
  vals = f(X, Y)
  signs = jnp.sign(vals)
  x_ids, y_ids = jnp.where(signs[:-1] * signs[1:] <= 0)
  starts = xs[x_ids]
  ends = xs[x_ids + 1]
  args = ys[y_ids]
  return starts, ends, args


def brentq_candidates(vmap_f, *values, args=()):
  # change the position of meshgrid values
  values = tuple((v.value if isinstance(v, bm.JaxArray) else v) for v in values)
  xs = values[0]
  mesh_values = jnp.meshgrid(*values)
  if bm.ndim(mesh_values[0]) > 1:
    mesh_values = tuple(jnp.moveaxis(m, 0, 1) for m in mesh_values)
  mesh_values = tuple(m.flatten() for m in mesh_values)
  # function outputs
  signs = jnp.sign(vmap_f(*(mesh_values + args)))
  # compute the selected values
  signs = signs.reshape((xs.shape[0], -1))
  par_len = signs.shape[1]
  signs1 = signs.at[-1].set(1)  # discard the final row
  signs2 = jnp.vstack((signs[1:], signs[:1])).at[-1].set(1)  # discard the first row
  ids = jnp.where((signs1 * signs2).flatten() <= 0)[0]
  x_starts = mesh_values[0][ids]
  x_ends = mesh_values[0][ids + par_len]
  other_vals = tuple(v[ids] for v in mesh_values[1:])
  return x_starts, x_ends, other_vals


def brentq_roots(f, starts, ends, *vmap_args, args=()):
  in_axes = (0, 0, tuple([0] * len(vmap_args)) + tuple([None] * len(args)))
  vmap_f_opt = bm.jit(vmap(jax_brentq(f), in_axes=in_axes))
  all_args = vmap_args + args
  if len(all_args):
    res = vmap_f_opt(starts, ends, all_args)
  else:
    res = vmap_f_opt(starts, ends, )
  valid_idx = jnp.where(res['status'] == ECONVERGED)[0]
  roots = res['root'][valid_idx]
  vmap_args = tuple(a[valid_idx] for a in vmap_args)
  return roots, vmap_args


def brentq_roots2(vmap_f, starts, ends, *vmap_args, args=()):
  all_args = vmap_args + args
  res = vmap_f(starts, ends, all_args)
  valid_idx = jnp.where(res['status'] == ECONVERGED)[0]
  roots = res['root'][valid_idx]
  vmap_args = tuple(a[valid_idx] for a in vmap_args)
  return roots, vmap_args


def scipy_minimize_with_jax(fun, x0,
                            method=None,
                            args=(),
                            bounds=None,
                            constraints=(),
                            tol=None,
                            callback=None,
                            options=None):
  """
  A simple wrapper for scipy.optimize.minimize using JAX.

  Parameters
  ----------
  fun: function
    The objective function to be minimized, written in JAX code
    so that it is automatically differentiable.  It is of type,
    ```fun: x, *args -> float``` where `x` is a PyTree and args
    is a tuple of the fixed parameters needed to completely specify the function.

  x0: jnp.ndarray
    Initial guess represented as a JAX PyTree.

  args: tuple, optional.
    Extra arguments passed to the objective function
    and its derivative.  Must consist of valid JAX types; e.g. the leaves
    of the PyTree must be floats.

  method : str or callable, optional
    Type of solver.  Should be one of
        - 'Nelder-Mead' :ref:`(see here) <optimize.minimize-neldermead>`
        - 'Powell'      :ref:`(see here) <optimize.minimize-powell>`
        - 'CG'          :ref:`(see here) <optimize.minimize-cg>`
        - 'BFGS'        :ref:`(see here) <optimize.minimize-bfgs>`
        - 'Newton-CG'   :ref:`(see here) <optimize.minimize-newtoncg>`
        - 'L-BFGS-B'    :ref:`(see here) <optimize.minimize-lbfgsb>`
        - 'TNC'         :ref:`(see here) <optimize.minimize-tnc>`
        - 'COBYLA'      :ref:`(see here) <optimize.minimize-cobyla>`
        - 'SLSQP'       :ref:`(see here) <optimize.minimize-slsqp>`
        - 'trust-constr':ref:`(see here) <optimize.minimize-trustconstr>`
        - 'dogleg'      :ref:`(see here) <optimize.minimize-dogleg>`
        - 'trust-ncg'   :ref:`(see here) <optimize.minimize-trustncg>`
        - 'trust-exact' :ref:`(see here) <optimize.minimize-trustexact>`
        - 'trust-krylov' :ref:`(see here) <optimize.minimize-trustkrylov>`
        - custom - a callable object (added in version 0.14.0),
          see below for description.
    If not given, chosen to be one of ``BFGS``, ``L-BFGS-B``, ``SLSQP``,
    depending on if the problem has constraints or bounds.

  bounds : sequence or `Bounds`, optional
    Bounds on variables for L-BFGS-B, TNC, SLSQP, Powell, and
    trust-constr methods. There are two ways to specify the bounds:
        1. Instance of `Bounds` class.
        2. Sequence of ``(min, max)`` pairs for each element in `x`. None
        is used to specify no bound.
    Note that in order to use `bounds` you will need to manually flatten
    them in the same order as your inputs `x0`.

  constraints : {Constraint, dict} or List of {Constraint, dict}, optional
    Constraints definition (only for COBYLA, SLSQP and trust-constr).
    Constraints for 'trust-constr' are defined as a single object or a
    list of objects specifying constraints to the optimization problem.
    Available constraints are:
        - `LinearConstraint`
        - `NonlinearConstraint`
    Constraints for COBYLA, SLSQP are defined as a list of dictionaries.
    Each dictionary with fields:
        type : str
            Constraint type: 'eq' for equality, 'ineq' for inequality.
        fun : callable
            The function defining the constraint.
        jac : callable, optional
            The Jacobian of `fun` (only for SLSQP).
        args : sequence, optional
            Extra arguments to be passed to the function and Jacobian.
    Equality constraint means that the constraint function result is to
    be zero whereas inequality means that it is to be non-negative.
    Note that COBYLA only supports inequality constraints.

    Note that in order to use `constraints` you will need to manually flatten
    them in the same order as your inputs `x0`.

  tol : float, optional
    Tolerance for termination. For detailed control, use solver-specific
    options.

  options : dict, optional
      A dictionary of solver options. All methods accept the following
      generic options:
          maxiter : int
              Maximum number of iterations to perform. Depending on the
              method each iteration may use several function evaluations.
          disp : bool
              Set to True to print convergence messages.
      For method-specific options, see :func:`show_options()`.

  callback : callable, optional
      Called after each iteration. For 'trust-constr' it is a callable with
      the signature:
          ``callback(xk, OptimizeResult state) -> bool``
      where ``xk`` is the current parameter vector represented as a PyTree,
       and ``state`` is an `OptimizeResult` object, with the same fields
      as the ones from the return. If callback returns True the algorithm
      execution is terminated.

      For all the other methods, the signature is:
          ```callback(xk)```
      where `xk` is the current parameter vector, represented as a PyTree.

  Returns
  -------
  res : The optimization result represented as a ``OptimizeResult`` object.
    Important attributes are:
        ``x``: the solution array, represented as a JAX PyTree
        ``success``: a Boolean flag indicating if the optimizer exited successfully
        ``message``: describes the cause of the termination.
    See `scipy.optimize.OptimizeResult` for a description of other attributes.

  """
  if soptimize is None:
    raise errors.PackageMissingError(f'"scipy" must be installed when user want to use '
                                     f'function: {scipy_minimize_with_jax}')

  # Use tree flatten and unflatten to convert params x0 from PyTrees to flat arrays
  x0_flat, unravel = ravel_pytree(x0)

  # Wrap the objective function to consume flat _original_
  # numpy arrays and produce scalar outputs.
  def fun_wrapper(x_flat, *args):
    x = unravel(x_flat)
    r = fun(x, *args)
    r = r.value if isinstance(r, bm.JaxArray) else r
    return float(r)

  # Wrap the gradient in a similar manner
  jac = jit(grad(fun))

  def jac_wrapper(x_flat, *args):
    x = unravel(x_flat)
    g_flat, _ = ravel_pytree(jac(x, *args))
    return np.array(g_flat)

  # Wrap the callback to consume a pytree
  def callback_wrapper(x_flat, *args):
    if callback is not None:
      x = unravel(x_flat)
      return callback(x, *args)

  # Minimize with scipy
  results = soptimize.minimize(fun_wrapper,
                                    x0_flat,
                                    args=args,
                                    method=method,
                                    jac=jac_wrapper,
                                    callback=callback_wrapper,
                                    bounds=bounds,
                                    constraints=constraints,
                                    tol=tol,
                                    options=options)

  # pack the output back into a PyTree
  results["x"] = unravel(results["x"])
  return results


def roots_of_1d_by_x(f, candidates, args=()):
  """Find the roots of the given function by numerical methods.
  """
  f = f_without_jaxarray_return(f)
  candidates = candidates.value if isinstance(candidates, bm.JaxArray) else candidates
  args = tuple(a.value if isinstance(candidates, bm.JaxArray) else a for a in args)
  vals = f(candidates, *args)
  signs = jnp.sign(vals)
  zero_sign_idx = jnp.where(signs == 0)[0]
  fps = candidates[zero_sign_idx]
  candidate_ids = jnp.where(signs[:-1] * signs[1:] < 0)[0]
  if len(candidate_ids) <= 0:
    return fps
  starts = candidates[candidate_ids]
  ends = candidates[candidate_ids + 1]
  f_opt = bm.jit(vmap(jax_brentq(f), in_axes=(0, 0, None)))
  res = f_opt(starts, ends, args)
  valid_idx = jnp.where(res['status'] == ECONVERGED)[0]
  fps2 = res['root'][valid_idx]
  return jnp.concatenate([fps, fps2])


def roots_of_1d_by_xy(f, starts, ends, args):
  f = f_without_jaxarray_return(f)
  f_opt = bm.jit(vmap(jax_brentq(f)))
  res = f_opt(starts, ends, (args,))
  valid_idx = jnp.where(res['status'] == ECONVERGED)[0]
  xs = res['root'][valid_idx]
  ys = args[valid_idx]
  return xs, ys


# @tools.numba_jit
def numpy_brentq(f, a, b, args=(), xtol=2e-14, maxiter=200, rtol=4 * np.finfo(float).eps):
  """
  Find a root of a function in a bracketing interval using Brent's method
  adapted from Scipy's brentq.

  Uses the classic Brent's method to find a zero of the function `f` on
  the sign changing interval [a , b].

  Parameters
  ----------
  f : callable
      Python function returning a number.  `f` must be continuous.
  a : number
      One end of the bracketing interval [a,b].
  b : number
      The other end of the bracketing interval [a,b].
  args : tuple, optional(default=())
      Extra arguments to be used in the function call.
  xtol : number, optional(default=2e-12)
      The computed root ``x0`` will satisfy ``np.allclose(x, x0,
      atol=xtol, rtol=rtol)``, where ``x`` is the exact root. The
      parameter must be nonnegative.
  rtol : number, optional(default=4*np.finfo(float).eps)
      The computed root ``x0`` will satisfy ``np.allclose(x, x0,
      atol=xtol, rtol=rtol)``, where ``x`` is the exact root.
  maxiter : number, optional(default=100)
      Maximum number of iterations.
  """
  if xtol <= 0:
    raise ValueError("xtol is too small (<= 0)")
  if maxiter < 1:
    raise ValueError("maxiter must be greater than 0")

  # Convert to float
  xpre = a * 1.0
  xcur = b * 1.0

  # Conditional checks for intervals in methods involving bisection
  fpre = f(xpre, *args)
  fcur = f(xcur, *args)
  funcalls = 2

  if fpre * fcur > 0:
    raise ValueError("f(a) and f(b) must have different signs")
  root = 0.0
  status = ECONVERR

  # Root found at either end of [a,b]
  if fpre == 0:
    root = xpre
    status = ECONVERGED
  if fcur == 0:
    root = xcur
    status = ECONVERGED

  root, status = root, status

  # Check for sign error and early termination
  if status == ECONVERGED:
    itr = 0
  else:
    # Perform Brent's method
    for itr in range(maxiter):
      if fpre * fcur < 0:
        xblk = xpre
        fblk = fpre
        spre = scur = xcur - xpre
      if abs(fblk) < abs(fcur):
        xpre = xcur
        xcur = xblk
        xblk = xpre

        fpre = fcur
        fcur = fblk
        fblk = fpre

      delta = (xtol + rtol * abs(xcur)) / 2
      sbis = (xblk - xcur) / 2

      # Root found
      if fcur == 0 or abs(sbis) < delta:
        status = ECONVERGED
        root = xcur
        itr += 1
        break

      if abs(spre) > delta and abs(fcur) < abs(fpre):
        if xpre == xblk:
          # interpolate
          stry = -fcur * (xcur - xpre) / (fcur - fpre)
        else:
          # extrapolate
          dpre = (fpre - fcur) / (xpre - xcur)
          dblk = (fblk - fcur) / (xblk - xcur)
          stry = -fcur * (fblk * dblk - fpre * dpre) / \
                 (dblk * dpre * (fblk - fpre))

        if 2 * abs(stry) < min(abs(spre), 3 * abs(sbis) - delta):
          # good short step
          spre = scur
          scur = stry
        else:
          # bisect
          spre = sbis
          scur = sbis
      else:
        # bisect
        spre = sbis
        scur = sbis

      xpre = xcur
      fpre = fcur
      if abs(scur) > delta:
        xcur += scur
      else:
        xcur += (delta if sbis > 0 else -delta)
      fcur = f(xcur, *args)
      funcalls += 1

  if status == ECONVERR:
    raise RuntimeError("Failed to converge")

  # x, funcalls, iterations = root, funcalls, itr
  return root, funcalls, itr


# @tools.numba_jit
def find_root_of_1d_numpy(f, f_points, args=(), tol=1e-8):
  """Find the roots of the given function by numerical methods.

  Parameters
  ----------
  f : callable
      The function.
  f_points : np.ndarray, list, tuple
      The value points.

  Returns
  -------
  roots : list
      The roots.
  """
  vals = f(f_points, *args)
  fs_len = len(f_points)
  signs = np.sign(vals)

  roots = []
  sign_l = signs[0]
  point_l = f_points[0]
  idx = 1
  while idx < fs_len and sign_l == 0.:
    roots.append(f_points[idx - 1])
    sign_l = signs[idx]
    idx += 1
  while idx < fs_len:
    sign_r = signs[idx]
    point_r = f_points[idx]
    if sign_r == 0.:
      roots.append(point_r)
      if idx + 1 < fs_len:
        sign_l = sign_r
        point_l = point_r
      else:
        break
      idx += 1
    else:
      if not np.isnan(sign_r) and sign_l != sign_r:
        root, funcalls, itr = numpy_brentq(f, point_l, point_r, args)
        if abs(f(root, *args)) < tol: roots.append(root)
      sign_l = sign_r
      point_l = point_r
      idx += 1

  return roots



# -*- coding: utf-8 -*-

import numpy as np

from brainpy import math as bm, tools
from .base import InterLayerInitializer

__all__ = [
  'Normal',
  'Uniform',
  'VarianceScaling',
  'KaimingUniform',
  'KaimingNormal',
  'XavierUniform',
  'XavierNormal',
  'LecunUniform',
  'LecunNormal',
  'Orthogonal',
  'DeltaOrthogonal',
]


def _compute_fans(shape, in_axis=-2, out_axis=-1):
  receptive_field_size = np.prod(shape) / shape[in_axis] / shape[out_axis]
  fan_in = shape[in_axis] * receptive_field_size
  fan_out = shape[out_axis] * receptive_field_size
  return fan_in, fan_out


class Normal(InterLayerInitializer):
  """Initialize weights with normal distribution.

  Parameters
  ----------
  scale : float
    The gain of the derivation of the normal distribution.

  """

  def __init__(self, scale=1., seed=None):
    super(Normal, self).__init__()
    self.scale = scale
    self.seed = seed
    self.rng = np.random.RandomState(seed=seed)

  def __call__(self, shape, dtype=None):
    shape = [tools.size2num(d) for d in shape]
    weights = self.rng.normal(size=shape, scale=self.scale)
    return bm.asarray(weights, dtype=dtype)

  def __repr__(self):
    return f'{self.__class__.__name__}(scale={self.scale}, seed={self.seed})'


class Uniform(InterLayerInitializer):
  """Initialize weights with uniform distribution.

  Parameters
  ----------
  min_val : float
    The lower limit of the uniform distribution.
  max_val : float
    The upper limit of the uniform distribution.

  """

  def __init__(self, min_val=0., max_val=1., seed=None):
    super(Uniform, self).__init__()
    self.min_val = min_val
    self.max_val = max_val
    self.seed = seed
    self.rng = np.random.RandomState(seed=seed)

  def __call__(self, shape, dtype=None):
    shape = [tools.size2num(d) for d in shape]
    r = self.rng.uniform(low=self.min_val, high=self.max_val, size=shape)
    return bm.asarray(r, dtype=dtype)

  def __repr__(self):
    return (f'{self.__class__.__name__}(min_val={self.min_val}, '
            f'max_val={self.max_val}, seed={self.seed})')


class VarianceScaling(InterLayerInitializer):
  def __init__(self, scale, mode, distribution, in_axis=-2, out_axis=-1, seed=None):
    self.scale = scale
    self.mode = mode
    self.in_axis = in_axis
    self.out_axis = out_axis
    self.distribution = distribution
    self.seed = seed
    self.rng = np.random.RandomState(seed=seed)

  def __call__(self, shape, dtype=None):
    shape = [tools.size2num(d) for d in shape]
    fan_in, fan_out = _compute_fans(shape, in_axis=self.in_axis, out_axis=self.out_axis)
    if self.mode == "fan_in":
      denominator = fan_in
    elif self.mode == "fan_out":
      denominator = fan_out
    elif self.mode == "fan_avg":
      denominator = (fan_in + fan_out) / 2
    else:
      raise ValueError("invalid mode for variance scaling initializer: {}".format(self.mode))
    variance = bm.array(self.scale / denominator, dtype=dtype)
    if self.distribution == "truncated_normal":
      from scipy.stats import truncnorm
      # constant is stddev of standard normal truncated to (-2, 2)
      stddev = bm.sqrt(variance) / bm.array(.87962566103423978, dtype)
      res = truncnorm(-2, 2).rvs(shape) * stddev
    elif self.distribution == "normal":
      res = self.rng.normal(size=shape) * bm.sqrt(variance)
    elif self.distribution == "uniform":
      res = self.rng.uniform(low=-1, high=1, size=shape) * bm.sqrt(3 * variance)
    else:
      raise ValueError("invalid distribution for variance scaling initializer")
    return bm.asarray(res, dtype=dtype)

  def __repr__(self):
    name = self.__class__.__name__
    blank = ' ' * len(name)
    return (f'{name}(scale={self.scale}, mode={self.mode}, in_axis={self.in_axis}, \n'
            f'{blank}out_axis={self.out_axis}, distribution={self.distribution}, seed={self.seed})')


class KaimingUniform(VarianceScaling):
  def __init__(self, scale=2.0, mode="fan_in",
               distribution="uniform",
               in_axis=-2, out_axis=-1,
               seed=None):
    super(KaimingUniform, self).__init__(scale, mode, distribution,
                                         in_axis=in_axis, out_axis=out_axis,
                                         seed=seed)


class KaimingNormal(VarianceScaling):
  def __init__(self, scale=2.0, mode="fan_in",
               distribution="truncated_normal",
               in_axis=-2, out_axis=-1,
               seed=None):
    super(KaimingNormal, self).__init__(scale, mode, distribution,
                                        in_axis=in_axis, out_axis=out_axis,
                                        seed=seed)


class XavierUniform(VarianceScaling):
  def __init__(self, scale=1.0, mode="fan_avg",
               distribution="uniform",
               in_axis=-2, out_axis=-1,
               seed=None):
    super(XavierUniform, self).__init__(scale, mode, distribution,
                                        in_axis=in_axis, out_axis=out_axis,
                                        seed=seed)


class XavierNormal(VarianceScaling):
  def __init__(self, scale=1.0, mode="fan_avg",
               distribution="truncated_normal",
               in_axis=-2, out_axis=-1,
               seed=None):
    super(XavierNormal, self).__init__(scale, mode, distribution,
                                       in_axis=in_axis, out_axis=out_axis,
                                       seed=seed)


class LecunUniform(VarianceScaling):
  def __init__(self, scale=1.0, mode="fan_in",
               distribution="uniform",
               in_axis=-2, out_axis=-1,
               seed=None):
    super(LecunUniform, self).__init__(scale, mode, distribution,
                                       in_axis=in_axis, out_axis=out_axis,
                                       seed=seed)


class LecunNormal(VarianceScaling):
  def __init__(self, scale=1.0, mode="fan_in",
               distribution="truncated_normal",
               in_axis=-2, out_axis=-1,
               seed=None):
    super(LecunNormal, self).__init__(scale, mode, distribution,
                                      in_axis=in_axis, out_axis=out_axis,
                                      seed=seed)


class Orthogonal(InterLayerInitializer):
  """
  Construct an initializer for uniformly distributed orthogonal matrices.

  If the shape is not square, the matrix will have orthonormal rows or columns
  depending on which side is smaller.
  """

  def __init__(self, scale=1., axis=-1, seed=None):
    super(Orthogonal, self).__init__()
    self.scale = scale
    self.axis = axis
    self.seed = seed
    self.rng = np.random.RandomState(seed=seed)

  def __call__(self, shape, dtype=None):
    shape = [tools.size2num(d) for d in shape]
    n_rows = shape[self.axis]
    n_cols = np.prod(shape) // n_rows
    matrix_shape = (n_rows, n_cols) if n_rows > n_cols else (n_cols, n_rows)
    norm_dst = self.rng.normal(size=matrix_shape)
    q_mat, r_mat = np.linalg.qr(norm_dst)
    # Enforce Q is uniformly distributed
    q_mat *= np.sign(np.diag(r_mat))
    if n_rows < n_cols:
      q_mat = q_mat.T
    q_mat = np.reshape(q_mat, (n_rows,) + tuple(np.delete(shape, self.axis)))
    q_mat = np.moveaxis(q_mat, 0, self.axis)
    return self.scale * bm.asarray(q_mat, dtype=dtype)

  def __repr__(self):
    return f'{self.__class__.__name__}(scale={self.scale}, axis={self.axis}, seed={self.seed})'


class DeltaOrthogonal(InterLayerInitializer):
  """
  Construct an initializer for delta orthogonal kernels; see arXiv:1806.05393.

  The shape must be 3D, 4D or 5D.
  """

  def __init__(self, scale=1.0, axis=-1, ):
    super(DeltaOrthogonal, self).__init__()
    self.scale = scale
    self.axis = axis

  def __call__(self, shape, dtype=None):
    shape = [tools.size2num(d) for d in shape]
    if len(shape) not in [3, 4, 5]:
      raise ValueError("Delta orthogonal initializer requires a 3D, 4D or 5D shape.")
    if shape[-1] < shape[-2]:
      raise ValueError("`fan_in` must be less or equal than `fan_out`. ")
    ortho_init = Orthogonal(scale=self.scale, axis=self.axis)
    ortho_matrix = ortho_init(shape[-2:], dtype=dtype)
    W = bm.zeros(shape, dtype=dtype)
    if len(shape) == 3:
      k = shape[0]
      W[(k - 1) // 2, ...] = ortho_matrix
    elif len(shape) == 4:
      k1, k2 = shape[:2]
      W[(k1 - 1) // 2, (k2 - 1) // 2, ...] = ortho_matrix
    else:
      k1, k2, k3 = shape[:3]
      W[(k1 - 1) // 2, (k2 - 1) // 2, (k3 - 1) // 2, ...] = ortho_matrix
    return W

  def __repr__(self):
    return f'{self.__class__.__name__}(scale={self.scale}, axis={self.axis})'



# -*- coding: utf-8 -*-

import jax.numpy as jnp
from jax import vmap
import numpy as np

import brainpy.math as bm
from .function import f_without_jaxarray_return
from .measurement import euclidean_distance

__all__ = [
  'Segment',
  'check_initials',
  'check_plot_durations',
  'get_sign',
  'get_sign2',
  'keep_unique',
  'rescale',
]


class Segment(object):
  def __init__(self, targets, num_segments):
    assert isinstance(targets, (tuple, list))
    # num segments
    if isinstance(num_segments, int):
      num_segments = tuple([num_segments] * len(targets))
    assert isinstance(num_segments, (tuple, list)) and len(num_segments) == len(targets)
    arg_lens = tuple(len(p) for p in targets)
    self. arg_pre_len = tuple(int(np.ceil(l / num_segments[i])) for i, l in enumerate(arg_lens))
    arg_id_segments = tuple(np.arange(0, l, self. arg_pre_len[i]) for i, l in enumerate(arg_lens))
    self. arg_id_segments = tuple(ids.flatten() for ids in np.meshgrid(*arg_id_segments))
    if len(arg_id_segments) == 0:
     self. arg_id_segments = ((0,),)
    self.targets = targets

  def __iter__(self):
    for ids in zip(*self. arg_id_segments):
      yield tuple(p[ids[i]: ids[i] + self. arg_pre_len[i]] for i, p in enumerate(self.targets))


def check_initials(initials, target_var_names):
  # check the initial values
  assert isinstance(initials, dict)
  for p in target_var_names:
    assert p in initials
  initials = {p: bm.asarray(initials[p], dtype=bm.float_) for p in target_var_names}
  len_of_init = []
  for v in initials.values():
    assert isinstance(v, (tuple, list, np.ndarray, jnp.ndarray, bm.ndarray))
    len_of_init.append(len(v))
  len_of_init = np.unique(len_of_init)
  assert len(len_of_init) == 1
  return initials


def check_plot_durations(plot_durations, duration, initials):
  if plot_durations is None:
    plot_durations = [(0., duration) for _ in range(len(initials))]
  if isinstance(plot_durations[0], (int, float)):
    assert len(plot_durations) == 2
    plot_durations = [plot_durations for _ in range(len(initials))]
  else:
    assert len(plot_durations) == len(initials)
    for dur in plot_durations:
      assert len(dur) == 2
  return plot_durations


def get_sign(f, xs, ys):
  f = f_without_jaxarray_return(f)
  xs = xs.value if isinstance(xs, bm.JaxArray) else xs
  ys = ys.value if isinstance(ys, bm.JaxArray) else ys
  Y, X = jnp.meshgrid(ys, xs)
  return jnp.sign(f(X, Y))


def get_sign2(f, *xyz, args=()):
  in_axes = tuple(range(len(xyz))) + tuple([None] * len(args))
  f = bm.jit(vmap(f_without_jaxarray_return(f), in_axes=in_axes))
  xyz = tuple((v.value if isinstance(v, bm.JaxArray) else v) for v in xyz)
  XYZ = jnp.meshgrid(*xyz)
  XYZ = tuple(jnp.moveaxis(v, 1, 0).flatten() for v in XYZ)
  shape = (len(v) for v in xyz)
  return jnp.sign(f(*(XYZ + args))).reshape(shape)


def keep_unique(candidates, tolerance=2.5e-2):
  """Filter unique fixed points by choosing a representative within tolerance.

  Parameters
  ----------
  candidates: np.ndarray
    The fixed points with the shape of (num_point, num_dim).

  Returns
  -------
  fps_and_ids : tuple
    A 2-tuple of (kept fixed points, ids of kept fixed points).
  """
  keep_ids = np.arange(candidates.shape[0])
  if tolerance <= 0.0:
    return candidates, keep_ids
  if candidates.shape[0] <= 1:
    return candidates, keep_ids

  # If point A and point B are within identical_tol of each other, and the
  # A is first in the list, we keep A.
  nfps = candidates.shape[0]
  distances = euclidean_distance(candidates)
  example_idxs = np.arange(nfps)
  all_drop_idxs = []
  for fidx in range(nfps - 1):
    distances_f = distances[fidx, fidx + 1:]
    drop_idxs = example_idxs[fidx + 1:][distances_f <= tolerance]
    all_drop_idxs += list(drop_idxs)
  keep_ids = np.setdiff1d(example_idxs, np.unique(all_drop_idxs))
  if keep_ids.shape[0] > 0:
    unique_fps = candidates[keep_ids, :]
  else:
    unique_fps = np.array([], dtype=candidates.dtype)
  return unique_fps, keep_ids


def rescale(min_max, scale=0.01):
  """Rescale lim."""
  min_, max_ = min_max
  length = max_ - min_
  min_ -= scale * length
  max_ += scale * length
  return min_, max_



# -*- coding: utf-8 -*-

import inspect

import brainpy.math as bm
from brainpy import errors

__all__ = [
  'f_without_jaxarray_return',
  'remove_return_shape',
  'get_args',
  'std_derivative',
  'std_func',
]


def f_without_jaxarray_return(f):
  def f2(*args, **kwargs):
    r = f(*args, **kwargs)
    return r.value if isinstance(r, bm.JaxArray) else r

  return f2


def remove_return_shape(f):
  def f2(*args, **kwargs):
    r = f(*args, **kwargs)
    if r.shape == (1,):r = r[0]
    return r
  return f2


def get_args(f, gather_var=True):
  reduced_args = []
  for name, par in inspect.signature(f).parameters.items():
    if par.kind is inspect.Parameter.POSITIONAL_OR_KEYWORD:
      reduced_args.append(par.name)
    elif par.kind is inspect.Parameter.KEYWORD_ONLY:
      reduced_args.append(par.name)
    elif par.kind is inspect.Parameter.VAR_POSITIONAL:
      raise errors.DiffEqError('Don not support positional only parameters, e.g., /')
    elif par.kind is inspect.Parameter.POSITIONAL_ONLY:
      raise errors.DiffEqError('Don not support positional only parameters, e.g., /')
    elif par.kind is inspect.Parameter.VAR_KEYWORD:
      raise errors.DiffEqError(f'Don not support dict of keyword arguments: {str(par)}')
    else:
      raise errors.DiffEqError(f'Unknown argument type: {par.kind}')

  if gather_var:
    var_names = []
    for a in reduced_args:
      if a == 't': break
      var_names.append(a)
    else:
      raise ValueError('Do not find time variable "t".')
    return var_names, reduced_args
  else:
    return reduced_args


def std_derivative(original_fargs, target_vars, target_pars):
  var = original_fargs[0]
  num_vars = len(target_vars)

  def inner(f):
    def call(*dyn_vars_and_pars, **fixed_vars_and_pars):
      params = dict()
      for i, v in enumerate(target_vars):
        if (v != var) and (v in original_fargs):
          params[v] = dyn_vars_and_pars[i]
      for j, p in enumerate(target_pars):
        if p in original_fargs:
          params[p] = dyn_vars_and_pars[num_vars + j]
      for k, v in fixed_vars_and_pars.items():
        if k in original_fargs:
          params[k] = v
      return f(dyn_vars_and_pars[target_vars.index(var)], 0., **params)

    return call

  return inner


def std_func(original_fargs, target_vars, target_pars):
  num_vars = len(target_vars)

  def inner(f):
    def call(*dyn_vars_and_pars, **fixed_vars_and_pars):
      params = dict()
      for i, v in enumerate(target_vars):
        if v in original_fargs:
          params[v] = dyn_vars_and_pars[i]
      for j, p in enumerate(target_pars):
        if p in original_fargs:
          params[p] = dyn_vars_and_pars[num_vars + j]
      for k, v in fixed_vars_and_pars.items():
        if k in original_fargs:
          params[k] = v
      return f(**params)

    return call

  return inner




# -*- coding: utf-8 -*-

"""
Dynamics simulation module.
"""


from .neurons import *
from .synapses import *
from .base import *
from .utils import *
from .runners import *

from . import neurons, synapses, utils, runners
